"""FoBiS.py main package"""

__version__ = "3.6.11"
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.

# from builtins import *



#!/usr/bin/env python
"""
Fetcher.py, module definition of FoBiS.py Fetcher object.

This module handles fetching and building GitHub-hosted FoBiS.py dependencies.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
import configparser
import os
import shutil
from collections.abc import Callable
from typing import Any, Optional

from .utils import print_fake, syswork


class Fetcher:
    """
    Fetcher handles fetching and building GitHub-hosted FoBiS.py dependencies.
    """

    DEPS_CONFIG_FILE = ".deps_config.ini"

    def __init__(
        self,
        deps_dir: str,
        print_n: Optional[Callable[..., None]] = None,
        print_w: Optional[Callable[..., None]] = None,
    ) -> None:
        """
        Parameters
        ----------
        deps_dir : str
          directory where dependencies are stored
        print_n : {None}
          function for printing normal messages
        print_w : {None}
          function for printing warning messages
        """
        self.deps_dir = deps_dir
        self.print_n = print_n if print_n is not None else print_fake
        self.print_w = print_w if print_w is not None else print_fake

    def parse_dep_spec(self, spec: str) -> dict[str, str]:
        """
        Parse a dependency spec string into a dict.

        Format: URL [:: branch=X] [:: tag=X] [:: rev=X] [:: mode=X]

        Parameters
        ----------
        spec : str
          dependency specification string

        Returns
        -------
        dict
          parsed fields: 'url', and optionally 'branch', 'tag', 'rev', 'mode'
        """
        parts = [p.strip() for p in spec.split("::")]
        result = {"url": parts[0].strip()}
        for part in parts[1:]:
            if "=" in part:
                key, _, value = part.partition("=")
                result[key.strip()] = value.strip()
        return result

    def fetch(
        self,
        name: str,
        url: str,
        branch: Optional[str] = None,
        tag: Optional[str] = None,
        rev: Optional[str] = None,
        update: bool = False,
    ) -> str:
        """
        Clone or update a dependency repo.

        Parameters
        ----------
        name : str
          dependency name (used as subdirectory name)
        url : str
          git repository URL
        branch : {None}
          branch to check out
        tag : {None}
          tag to check out
        rev : {None}
          revision/commit to check out
        update : {False}
          if True, run git fetch to update before checking out

        Returns
        -------
        str
          path to the cloned dependency directory
        """
        dep_dir = os.path.join(self.deps_dir, name)
        if not os.path.exists(self.deps_dir):
            os.makedirs(self.deps_dir)
        if not os.path.exists(dep_dir):
            self.print_n("Cloning " + name + " from " + url)
            result = syswork("git clone " + url + " " + dep_dir)
            if result[0] != 0:
                self.print_w("Error cloning " + name + ":\n" + result[1])
                return dep_dir
        elif update:
            self.print_n("Updating " + name)
            result = syswork("git -C " + dep_dir + " fetch")
            if result[0] != 0:
                self.print_w("Error fetching updates for " + name + ":\n" + result[1])
        else:
            self.print_n("Dependency " + name + " already fetched (use --update to refresh)")
        ref = tag or branch or rev
        if ref:
            self.print_n("Checking out " + ref + " for " + name)
            result = syswork("git -C " + dep_dir + " checkout " + ref)
            if result[0] != 0:
                self.print_w("Error checking out " + ref + " for " + name + ":\n" + result[1])
        elif update:
            result = syswork("git -C " + dep_dir + " merge --ff-only")
            if result[0] != 0:
                self.print_w("Error merging updates for " + name + ":\n" + result[1])
        return dep_dir

    def build_dep(self, name: str, dep_dir: str, mode: Optional[str] = None) -> None:
        """
        Build a dependency using FoBiS.py build.

        Parameters
        ----------
        name : str
          dependency name (for messages)
        dep_dir : str
          path to the dependency directory
        mode : {None}
          fobos mode to use for building
        """
        fobos_file = os.path.join(dep_dir, "fobos")
        if not os.path.exists(fobos_file):
            self.print_w('Error: dependency "' + name + '" has no fobos file in ' + dep_dir)
            return
        self.print_n("Building dependency " + name)
        old_pwd = os.getcwd()
        os.chdir(dep_dir)
        if mode:
            result = syswork("FoBiS.py build -mode " + mode)
        else:
            result = syswork("FoBiS.py build")
        os.chdir(old_pwd)
        if result[0] != 0:
            self.print_w("Error building " + name + ":\n" + result[1])
        else:
            self.print_n(result[1])

    def save_config(self, deps_info: list[dict[str, Any]]) -> None:
        """
        Write the deps config file for use by FoBiS.py build.

        Deps with use='fobos' are written as 'dependon' entries (library approach).
        Deps with use='sources' (default) are written as 'src' entries (direct source inclusion).

        Parameters
        ----------
        deps_info : list
          list of dicts with keys 'name', 'path', 'mode', 'use'
        """
        config = configparser.RawConfigParser()
        config.add_section("deps")
        dependon_entries = []
        src_entries = []
        for dep in deps_info:
            if dep.get("use", "sources") == "fobos":
                fobos_path = os.path.join(dep["path"], "fobos")
                entry = fobos_path + ":" + dep["mode"] if dep.get("mode") else fobos_path
                dependon_entries.append(entry)
            else:
                src_entries.append(dep["path"])
        if dependon_entries:
            config.set("deps", "dependon", " ".join(dependon_entries))
        if src_entries:
            config.set("deps", "src", " ".join(src_entries))
        config_path = os.path.join(self.deps_dir, self.DEPS_CONFIG_FILE)
        with open(config_path, "w") as cfg_file:
            config.write(cfg_file)
        self.print_n("Saved deps config to " + config_path)

    def _resolve_url(self, repo: str) -> str:
        """
        Convert a shorthand repo reference to a full GitHub HTTPS URL.

        Accepts 'user/repo' (GitHub shorthand) or any full URL unchanged.

        Parameters
        ----------
        repo : str
          repository reference

        Returns
        -------
        str
          full git-cloneable URL
        """
        if repo.startswith("http://") or repo.startswith("https://") or repo.startswith("git@"):
            return repo
        return "https://github.com/" + repo

    def install_from_github(
        self,
        repo: str,
        branch: Optional[str] = None,
        tag: Optional[str] = None,
        rev: Optional[str] = None,
        mode: Optional[str] = None,
        update: bool = False,
        no_build: bool = False,
        prefix: str = "./",
        bin_dir: str = "bin/",
        lib_dir: str = "lib/",
        include_dir: str = "include/",
    ) -> None:
        """
        Clone, build, and install a GitHub-hosted FoBiS project.

        Parameters
        ----------
        repo : str
          repository reference ('user/repo' shorthand or full URL)
        branch : {None}
          branch to check out
        tag : {None}
          tag to check out
        rev : {None}
          revision/commit to check out
        mode : {None}
          fobos mode to use when building
        update : {False}
          re-fetch before building
        no_build : {False}
          clone only, skip building and installing
        prefix : str
          installation prefix directory
        bin_dir : str
          sub-directory under prefix for executables
        lib_dir : str
          sub-directory under prefix for libraries
        include_dir : str
          sub-directory under prefix for module files
        """
        url = self._resolve_url(repo)
        name = url.rstrip("/").split("/")[-1]
        if name.endswith(".git"):
            name = name[:-4]
        dep_dir = self.fetch(name, url, branch=branch, tag=tag, rev=rev, update=update)
        if no_build:
            self.print_n("Cloned " + name + " to " + dep_dir + " (--no-build: skipping build and install)")
            return
        self._build_dep_tracked(name, dep_dir, mode=mode)
        self._install_artifacts(dep_dir, prefix, bin_dir, lib_dir, include_dir)

    def _build_dep_tracked(self, name: str, dep_dir: str, mode: Optional[str] = None) -> None:
        """
        Build a dependency using 'fobis build --track_build'.

        Parameters
        ----------
        name : str
          dependency name (for messages)
        dep_dir : str
          path to the dependency directory
        mode : {None}
          fobos mode to use for building
        """
        fobos_file = os.path.join(dep_dir, "fobos")
        if not os.path.exists(fobos_file):
            self.print_w('Error: "' + name + '" has no fobos file in ' + dep_dir)
            return
        self.print_n("Building " + name + " (with --track_build)")
        old_pwd = os.getcwd()
        os.chdir(dep_dir)
        cmd = "fobis build --track_build"
        if mode:
            cmd += " -mode " + mode
        result = syswork(cmd)
        os.chdir(old_pwd)
        if result[0] != 0:
            self.print_w("Error building " + name + ":\n" + result[1])
        else:
            self.print_n(result[1])

    def _install_artifacts(self, dep_dir: str, prefix: str, bin_dir: str, lib_dir: str, include_dir: str) -> None:
        """
        Scan dep_dir for .track_build files and copy artifacts to the prefix.

        Parameters
        ----------
        dep_dir : str
          root of the cloned dependency
        prefix : str
          installation prefix
        bin_dir : str
          sub-directory under prefix for executables
        lib_dir : str
          sub-directory under prefix for libraries
        include_dir : str
          sub-directory under prefix for module/include files
        """
        installed_any = False
        for root, _, files in os.walk(dep_dir):
            for filename in files:
                if not filename.endswith(".track_build"):
                    continue
                track_file = configparser.ConfigParser()
                track_file.read(os.path.join(root, filename))
                if not track_file.has_option("build", "output"):
                    continue
                output = track_file.get("build", "output")
                if not os.path.exists(output):
                    continue
                is_program = track_file.has_option("build", "program") and track_file.get("build", "program")
                is_library = track_file.has_option("build", "library") and track_file.get("build", "library")
                if is_program:
                    dest = os.path.join(prefix, bin_dir)
                    os.makedirs(dest, exist_ok=True)
                    self.print_n('Installing "' + output + '" -> "' + dest + '"')
                    shutil.copy(output, dest)
                    installed_any = True
                if is_library:
                    dest = os.path.join(prefix, lib_dir)
                    os.makedirs(dest, exist_ok=True)
                    self.print_n('Installing "' + output + '" -> "' + dest + '"')
                    shutil.copy(output, dest)
                    installed_any = True
                    if track_file.has_option("build", "mod_file"):
                        mod_file = track_file.get("build", "mod_file")
                        inc_dest = os.path.join(prefix, include_dir)
                        os.makedirs(inc_dest, exist_ok=True)
                        self.print_n('Installing "' + mod_file + '" -> "' + inc_dest + '"')
                        shutil.copy(mod_file, inc_dest)
        if not installed_any:
            self.print_w(
                "No installable artifacts found. Ensure the project fobos uses --track_build or check the fobos mode."
            )

    def load_config(self) -> dict[str, list[str]]:
        """
        Read the deps config file and return a dict with 'dependon' and 'src' lists.

        Returns
        -------
        dict
          'dependon' : list of fobos-path strings for use=fobos deps
          'src'      : list of directory paths for use=sources deps
          either key is absent when no entries of that type exist
        """
        config_path = os.path.join(self.deps_dir, self.DEPS_CONFIG_FILE)
        if not os.path.exists(config_path):
            return {}
        config = configparser.RawConfigParser()
        config.read(config_path)
        result = {}
        if config.has_option("deps", "dependon"):
            result["dependon"] = config.get("deps", "dependon").split()
        if config.has_option("deps", "src"):
            result["src"] = config.get("deps", "src").split()
        return result



"""
Gcov.py, module definition of Gcov class.

This is a class designed for analyzing gcov files.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from builtins import *
try:
    from pygooglechart import PieChart2D

    __graph__ = True
except ImportError:
    # print("Module 'pygooglechart' not found: Gcov graphs disabled")
    __graph__ = False
import os
import re
import sys
from operator import itemgetter

__str_kw_elemental__ = r"[Ee][Ll][Ee][Mm][Ee][Nn][Tt][Aa][Ll]"
__str_kw_pure__ = r"[Pp][Uu][Rr][Ee]"
__str_kw_subroutine__ = r"[Ss][Uu][Bb][Rr][Oo][Uu][Tt][Ii][Nn][Ee]"
__str_kw_function__ = r"[Ff][Uu][Nn][Cc][Tt][Ii][Oo][Nn]"
__str_name__ = r"(?P<name>[a-zA-Z][a-zA-Z0-9_]*)"
__str_procedure__ = (
    r"^(\s*)"  # initial white spaces
    + r"("
    + __str_kw_elemental__
    + r"|"
    + __str_kw_pure__
    + r")?\s+"  # "elemental" or "pure" attribute
    + r"(?P<ptype>"
    + __str_kw_function__
    + r"|"
    + __str_kw_subroutine__
    + r")\s+"  # type of procedure
    + r"(?P<pname>"
    + __str_name__
    + r"\s*).*"
)  # name of procedure
__regex_procedure__ = re.compile(__str_procedure__)


def _mermaid_pie(title, executed, unexecuted):
    """Return a Mermaid pie-chart fenced code block string with surrounding blank lines."""
    return (
        "```mermaid\npie showData\n"
        + "    title "
        + title
        + "\n"
        + '    "Executed" : '
        + executed
        + "\n"
        + '    "Unexecuted" : '
        + unexecuted
        + "\n"
        + "```\n\n"
    )


class Gcov:
    """Gcov is an object that handles gcov file, its attributes and methods."""

    def __init__(self, filename=None):
        """
        Parameters
        ----------
        filename : {None}
          string containing the gcov filename
        """
        self.filename = filename
        self.procedures = []
        self.coverage = []
        self.metrics = {"coverage": None, "procedures": None}
        self.l_pie_url = ""
        self.p_pie_url = ""
        return

    def _metrics(self):
        """
        Getg gcov metrics.
        """
        if len(self.coverage) > 0:
            lnumber = sum(1 for cov in self.coverage if isinstance(cov, int))
            elnumber = sum(cov > 0 for cov in self.coverage if isinstance(cov, int))
            unelnumber = sum(cov == 0 for cov in self.coverage if isinstance(cov, int))
            if lnumber > 0:
                elnumber_per = int(round(elnumber * 100.0 / lnumber))
                unelnumber_per = int(round(unelnumber * 100.0 / lnumber))
            else:
                elnumber_per = 0
                unelnumber_per = 0
            if elnumber > 0:
                ahits = sum(cov for cov in self.coverage if isinstance(cov, int)) / elnumber
            else:
                ahits = 0
            self.metrics["coverage"] = [
                str(lnumber),
                str(elnumber),
                str(unelnumber),
                str(elnumber_per),
                str(unelnumber_per),
                str(ahits),
            ]
        if len(self.procedures) > 0:
            pnumber = len(self.procedures)
            epnumber = sum(proc[2] > 0 for proc in self.procedures)
            unepnumber = sum(proc[2] == 0 for proc in self.procedures)
            if pnumber > 0:
                epnumber_per = int(round(epnumber * 100.0 / pnumber))
                unepnumber_per = int(round(unepnumber * 100.0 / pnumber))
            else:
                epnumber_per = 0
                unepnumber_per = 0
            if epnumber > 0:
                ahits = sum(proc[2] for proc in self.procedures) / epnumber
            else:
                ahits = 0
            self.metrics["procedures"] = [
                str(pnumber),
                str(epnumber),
                str(unepnumber),
                str(epnumber_per),
                str(unepnumber_per),
                str(ahits),
            ]
        return

    def _get_graphs_url(self):
        """
        Get pie-chart graphs urls from online google chart service.

        Returns
        ----------
        str, str
          string containing the url of online google pie-chart graphs
        """

        def pie(size, title, labels, legend, data, colors):
            """
            Function for creating a pie chart.

            Parameters
            ----------
            size : list
              width and height
            title : str
              title of the graph
            lables : list
              list of string containing labels of the data
            legend : list
              list of string containing legend of the data
            data : list
              data to be plotted
            colors : list
              list of string containing the colors of each data in PRGB format

            Returns
            -------
            PieChart2D object
            """
            graph = PieChart2D(size[0], size[1])
            graph.set_title(title)
            graph.set_legend(legend)
            graph.set_pie_labels(labels)
            graph.add_data(data)
            graph.set_colours(colors)
            return graph

        if __graph__:
            if self.metrics["coverage"]:
                l_pie = pie(
                    size=[300, 150],
                    title="Coverage of executable lines",
                    labels=[self.metrics["coverage"][3] + "%", self.metrics["coverage"][4] + "%"],
                    legend=["Executed", "Unexecuted"],
                    data=[int(self.metrics["coverage"][3]), int(self.metrics["coverage"][4])],
                    colors=["65C1FF", "FF9260"],
                )
                self.l_pie_url = "[![lines](" + l_pie.get_url() + ")]()"
            if self.metrics["procedures"]:
                p_pie = pie(
                    size=[300, 150],
                    title="Coverage of procedures",
                    labels=[self.metrics["procedures"][3] + "%", self.metrics["procedures"][4] + "%"],
                    legend=["Executed", "Unexecuted"],
                    data=[int(self.metrics["procedures"][3]), int(self.metrics["procedures"][4])],
                    colors=["65C1FF", "FF9260"],
                )
                self.p_pie_url = "[![procedures](" + p_pie.get_url() + ")]()"
        return

    def parse(self):
        """
        Parse gcov file.
        """

        def _coverage_update(ignoring, cov_num, line_num, text, coverage):
            """
            Update coverage list.
            """
            if re.search(r"\bLCOV_EXCL_START\b", text):
                if ignoring:
                    sys.stderr.write("Warning: %s:%d: nested LCOV_EXCL_START, please fix\n" % (self.filename, line_num))
                ignoring = True
            elif re.search(r"\bLCOV_EXCL_END\b", text):
                if not ignoring:
                    sys.stderr.write(
                        "Warning: %s:%d: LCOV_EXCL_END outside of "
                        "exclusion zone, please fix\n" % (self.filename, line_num)
                    )
                ignoring = False
            if cov_num == "-":
                coverage.append(None)
            elif cov_num == "#####":
                if (
                    ignoring
                    or text.lstrip().startswith(("inline", "static"))
                    or text.strip() == "}"
                    or re.search(r"\bLCOV_EXCL_LINE\b", text)
                ):
                    coverage.append(None)
                else:
                    coverage.append(0)
            elif cov_num == "=====":
                coverage.append(0)
            elif "*" in cov_num:
                coverage.append(int(cov_num.rstrip("*")))
            else:
                try:
                    cov_num_int = int(cov_num)
                except ValueError:
                    cov_num_int = 0
                coverage.append(cov_num_int)
            return ignoring

        ignoring = False
        if self.filename:
            if os.path.exists(self.filename):
                with open(self.filename) as source:
                    for line in source:
                        if ":" not in line:
                            break
                        report_fields = line.split(":")
                        cov_num = report_fields[0].strip()
                        line_num = int(report_fields[1].strip())
                        text = report_fields[2]
                        proc_matching = re.match(__regex_procedure__, text)
                        if line_num == 0:
                            continue
                        if proc_matching:
                            ptype = proc_matching.group("ptype").strip()
                            pname = proc_matching.group("pname").strip()
                            if cov_num != "#####" and cov_num != "-":
                                try:
                                    pcov = int(cov_num.strip("*"))
                                except ValueError:
                                    pcov = 0
                            else:
                                pcov = 0
                            if cov_num != "-":  # needed due to abstract iterfaces
                                self.procedures.append([ptype, pname, pcov, line_num])
                        ignoring = _coverage_update(
                            ignoring=ignoring, cov_num=cov_num, line_num=line_num, text=text, coverage=self.coverage
                        )
        self._metrics()
        self._get_graphs_url()
        return

    def save(self, report_format="markdown", output=None, graphs=False):
        """
        Save gcov report analysis.

        Parameters
        ----------
        report_format : {'markdown'}
          output format of saved report
        output : {None}
          string output file name
        graphs : {False}
          flag for including graphs
        """
        string = []
        if len(self.metrics) > 0:
            if report_format == "markdown":
                string.append(
                    "### Coverage analysis of *" + os.path.splitext(os.path.basename(self.filename))[0] + "*\n"
                )
                string.append("\n")
                if self.metrics["coverage"]:
                    string.append("|Lines| | |\n")
                    string.append("| --- | --- | --- |\n")
                    string.append("|Executable lines            |" + self.metrics["coverage"][0] + "| |\n")
                    string.append(
                        "|Executed lines              |"
                        + self.metrics["coverage"][1]
                        + "|"
                        + self.metrics["coverage"][3]
                        + "%|\n"
                    )
                    string.append(
                        "|Unexecuted lines            |"
                        + self.metrics["coverage"][2]
                        + "|"
                        + self.metrics["coverage"][4]
                        + "%|\n"
                    )
                    string.append("|Average hits / executed     |" + self.metrics["coverage"][5] + "| |\n")
                    string.append("\n")
                    string.append(
                        _mermaid_pie(
                            "Lines (" + self.metrics["coverage"][3] + "% covered)",
                            self.metrics["coverage"][1],
                            self.metrics["coverage"][2],
                        )
                    )
                if self.metrics["procedures"]:
                    string.append("|Procedures| | |\n")
                    string.append("| --- | --- | --- |\n")
                    string.append("|Total procedures            |" + self.metrics["procedures"][0] + "| |\n")
                    string.append(
                        "|Executed procedures         |"
                        + self.metrics["procedures"][1]
                        + "|"
                        + self.metrics["procedures"][3]
                        + "%|\n"
                    )
                    string.append(
                        "|Unexecuted procedures       |"
                        + self.metrics["procedures"][2]
                        + "|"
                        + self.metrics["procedures"][4]
                        + "%|\n"
                    )
                    string.append("|Average hits / executed     |" + self.metrics["procedures"][5] + "| |\n")
                    string.append("\n")
                    string.append(
                        _mermaid_pie(
                            "Procedures (" + self.metrics["procedures"][3] + "% covered)",
                            self.metrics["procedures"][1],
                            self.metrics["procedures"][2],
                        )
                    )
                if graphs and __graph__ and (self.metrics["coverage"] or self.metrics["procedures"]):
                    string.append("\n --- \n")
                    if self.metrics["coverage"]:
                        string.append(self.l_pie_url + "\n")
                    if self.metrics["procedures"]:
                        string.append(self.p_pie_url + "\n")
                string.append("\n#### Unexecuted procedures\n")
                string.append("\n")
                if any(proc[2] == 0 for proc in self.procedures):
                    for proc in sorted(self.procedures, key=itemgetter(0, 1)):
                        if proc[2] == 0:
                            string.append(" + *" + proc[0] + "* **" + proc[1] + "**, line " + str(proc[3]) + "\n")
                else:
                    string.append(" + *none*\n")
                string.append("\n#### Executed procedures\n")
                string.append("\n")
                if any(proc[2] > 0 for proc in self.procedures):
                    for proc in sorted(self.procedures, key=itemgetter(2), reverse=True):
                        if proc[2] > 0:
                            string.append(
                                " + *" + proc[0] + "* **" + proc[1] + "**: tested **" + str(proc[2]) + "** times\n"
                            )
                else:
                    string.append(" + *none*\n")
                string.append("\n --- \n Report generated by [FoBiS.py](https://github.com/szaghi/FoBiS)")
        if output:
            fout = output
        else:
            fout = os.path.splitext(os.path.basename(self.filename))[0] + ".gcov-report.md"
        with open(fout, "w") as out:
            out.writelines(string)
        return



"""
utils.py, module definition of FoBiS.py util functions.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from builtins import *
import contextlib
import os
import shlex
import subprocess
import sys


def print_fake(input_obj, end="\n"):
    """Rename print."""
    print(input_obj, end=end)


def syswork(cmd):
    """
    Function for executing system command 'cmd': for compiling and linking files.
    """
    error = 0
    args = shlex.split(cmd) if isinstance(cmd, str) else cmd
    try:
        output = subprocess.check_output(args, shell=False, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as err:
        error = err.returncode
        output = err.output
    output = str(output, encoding="UTF-8")
    return [error, str(output)]


def syswork_steps(steps):
    """
    Run a sequence of compile steps for a single source file.

    Each step is a (cmd, cleanup_path_or_None) tuple. Steps are run in order;
    execution stops on the first non-zero exit code. If cleanup_path is set on
    a step, the file is removed with os.remove() after that step completes
    (regardless of success/failure).

    Returns [error, combined_output] matching the syswork contract.
    """
    combined_output = ""
    for cmd, cleanup in steps:
        error, output = syswork(cmd)
        combined_output += output
        if cleanup:
            with contextlib.suppress(OSError):
                os.remove(cleanup)
        if error:
            return [error, combined_output]
    return [0, combined_output]


def traverse_recursive(parsed_file, path=None):
    """
    The function traverse_recursive performs a yield-recursive traversing of pfile direct dependencies.
    """
    if path is None:
        path = list()
    path.append(parsed_file)
    yield path
    for dep in parsed_file.pfile_dep:
        if dep != parsed_file:
            for path in traverse_recursive(dep, path):
                yield path
    if path:
        path.pop()


def unique_seq(seq):
    """
    The function unique_seq returns the input sequence removing duplicated elements but peserving the original order.
    """
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]


def module_is_in(pfiles, module):
    """
    Find the parsed file containing the desidered module.

    Parameters
    ----------
    pfiles : list
      list of parsed files
    module : str
      module name

    Returns
    -------
    file_name : str
      name of file containing the module
    number : int
      number of file containing the module
    """
    file_name = ""
    number = -1
    for fnum, parsed_file in enumerate(pfiles):
        if parsed_file.module:
            for module_name in parsed_file.module_names:
                if module_name.lower() == module.lower():
                    file_name = parsed_file.name
                    number = fnum
                    break
    return file_name, number


def include_is_in(pfiles, include):
    """
    Function finding the parsed file containing the desidered include-file.

    Parameters
    ----------
    pfiles : list
      list of parsed files
    include : str
      include-file name

    Returns
    -------
    file_name : str
      name of file containing the include-file
    number : int
      number of file containing the include-file
    """
    file_name = ""
    number = -1
    for fnum, parsed_file in enumerate(pfiles):
        if os.path.basename(parsed_file.name) == include:
            file_name = parsed_file.name
            number = fnum
            break
    return file_name, number


def dependency_hiearchy(builder, pfiles, print_w=None, force_compile=False):
    """
    Build parsed files hierarchy.

    Paramters
    ---------
    builder : Builder object
    pfiles : list
      list of ParsedFile objects
    print_w : function
      function for printing emphized warning message
    force_compile : {False}
      flag for forcing (re-)compiling of all dependency
    """
    if print_w is None:
        print_wm = print_fake
    else:
        print_wm = print_w

    # direct dependencies list used after for building indirect (complete) dependencies list
    for parsed_file in pfiles:
        parsed_file.pfile_dep = []
        for dep in parsed_file.dependencies:
            if dep.type == "module":
                dep.file, fnum = module_is_in(pfiles=pfiles, module=dep.name)
                if fnum > -1:
                    if pfiles[fnum] not in parsed_file.pfile_dep:
                        parsed_file.pfile_dep.append(pfiles[fnum])
                else:
                    # there is no source containg the searched module, try to find it into a precompiled mod file into eventually included paths
                    for dinc in builder.dinc:
                        for _, _, files in os.walk(dinc):
                            for filename in files:
                                if (
                                    os.path.splitext(os.path.basename(filename))[0].lower() == dep.name.lower()
                                    and os.path.splitext(os.path.basename(filename))[1].lower() == ".mod"
                                ):
                                    fnum = 0
                                    break
                            if fnum == 0:
                                break
                        if fnum == 0:
                            break
                    if fnum != 0:
                        print_wm(
                            "Attention: the file '"
                            + parsed_file.name
                            + "' depends on '"
                            + dep.name
                            + "' that is unreachable"
                        )
                        # sys.exit(1)
            if dep.type == "include":
                dep.file, fnum = include_is_in(pfiles=pfiles, include=dep.name)
                if fnum > -1:
                    if pfiles[fnum] not in parsed_file.pfile_dep:
                        pfiles[fnum].program = False
                        pfiles[fnum].module = False
                        pfiles[fnum].nomodlib = False
                        pfiles[fnum].include = True
                        parsed_file.pfile_dep.append(pfiles[fnum])
                        if os.path.dirname(pfiles[fnum].name) not in builder.dinc:
                            builder.dinc.append(os.path.dirname(pfiles[fnum].name))
                else:
                    print_wm(
                        "Attention: the file '"
                        + parsed_file.name
                        + "' depends on '"
                        + dep.name
                        + "' that is unreachable"
                    )
                    # sys.exit(1)
    # indirect dependency list
    for parsed_file in pfiles:
        parsed_file.create_pfile_dep_all()
    # using the just created hiearchy for checking which files must be (re-)compiled
    for parsed_file in pfiles:
        parsed_file.check_compile(obj_dir=builder.obj_dir, force_compile=force_compile)


def remove_other_main(builder, pfiles, myself):
    """
    The function remove_other_main removes all compiled objects of other program than the current target under building.
    """
    for parsed_file in pfiles:
        if parsed_file.program and parsed_file.name != myself.name:
            if os.path.exists(builder.obj_dir + parsed_file.basename + ".o"):
                os.remove(builder.obj_dir + parsed_file.basename + ".o")


def check_results(results, log=None, print_w=None):
    """
    Check the result of system commands exectution.

    Parameters
    ----------
    results : list
      list of exectuions results
    log : {False}
      bool for activate errors log saving
    print_w : {None}
      function for printing emphized warning message
    """
    if print_w is None:
        print_wm = print_fake
    else:
        print_wm = print_w

    if not all(v[1] == "" for v in results):
        for result in results:
            if result[1] != "":
                print_wm(result[1])
    if not all(v[0] == 0 for v in results):
        for result in results:
            if result[0] != 0:
                sys.stderr.write(result[1])
                if log:
                    with open(log, "a") as logerror:
                        logerror.writelines(result[1])
        sys.exit(1)


def safe_mkdir(directory, print_w=None):
    """
    Create directory via safe checkings.

    Parameters
    ----------
    directory: str
      path of the directory to be created
    print_w : {None}
      function for printing emphized warning message
    """
    if print_w is None:
        print_wm = print_fake
    else:
        print_wm = print_w

    if os.path.exists(directory):
        if not os.path.isdir(directory):
            print_wm('Error: cannot create directory "' + directory + '", a file with the same name already exists!')
            sys.exit(1)
    else:
        os.makedirs(directory)
    return



"""Tests for FoBiS.py inline doctest functionality."""

import pytest

from tests.helpers import run_doctest


@pytest.mark.parametrize("n", range(1, 4))
def test_doctest(n):
    """Test doctest scenario n."""
    assert run_doctest(f"doctest-test{n}"), f"doctest-test{n} failed"



"""
Dependency.py, module definition of Dependency class.

This is a class designed for handling file dependency.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from builtins import *
import os


class Dependency:
    """Dependency is an object that handles a single file dependency, its attributes and methods."""

    def __init__(self, dtype="", name="", dfile=""):
        """
        Parameters
        ----------
        dtype : {""}
          type of dependency: "module" or "include" type
        name : {""}
          name of dependency: module name for "use" type or file name for include type
        file : {""}
          file name containing module in the case of "use" type

        Attributes
        ----------
        """
        self.type = dtype
        self.name = name
        self.file = dfile

    def __str__(self):
        string = []
        string.append("\n  Type: " + str(self.type))
        string.append("\n  Name: " + str(self.name))
        string.append("\n  File: " + str(self.file))
        return "".join(string)

    def printf(self):
        """Method for printing dependency data."""
        print(self)
        return

    def exist(self):
        """Method for checking the existance of a dependency file."""
        return os.path.exists(self.file)



"""
Builder.py, module definition of Builder class.
This is a class designed for controlling the building phase.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from builtins import *
try:
    from multiprocessing import Pool

    __parallel__ = True
except ImportError:
    print("Module 'multiprocessing' not found: parallel compilation disabled")
    __parallel__ = False
import operator
import os
from collections.abc import Callable
from typing import Any, Optional

from .Compiler import Compiler
from .utils import check_results, print_fake, safe_mkdir, syswork, syswork_steps


class Builder:
    """Builder is an object that handles the building system, its attributes and methods."""

    def __init__(
        self,
        cliargs: Any,
        print_n: Optional[Callable[..., None]] = None,
        print_w: Optional[Callable[..., None]] = None,
    ) -> None:
        """
        Parameters
        ----------
        cliargs : argparse object
        print_n : {None}
          function for printing normal message
        print_w : {None}
          function for printing emphized warning message

        Attributes
        ----------
        compiler : Compiler object
          compiler used
        src_dir = {"./"}
          directory into which search source files
        build_dir : {"./"}
          directory containing built files
        obj_dir : {"./"}
          directory containing compiled object files
        mod_dir : {"./"}
          directory containing .mod files
        lib_dir : {None}
          list of directories searched for libraries
        dinc : {None}
          list of directories for searching included files
        libs : {None}
          list of external libraries that are not into the path: must be passed with full paths
        vlibs : {None}
          list of external libraries that are not into the path and that are volatile (can changed thus triggering re-building): must be passed with full paths
        ext_libs : {None}
          list of external libraries that are into the path: are linked as '-llibrary.[a|so]' and searched into '-Ldir'
        ext_vlibs : {None}
          list of external libraries that are into the path and that are volatile (can changed thus triggering re-building): are linked as '-llibrary.[a|so]' and searched into '-Ldir'
        preprocessor : {None}
          activate preprocessor
        preprocessor_dir : {None}
          directory containing sources processed by preprocessor (if any); by default are removed after used
        preprocessor_ext : {None}
          list of file extensions to be processed by preprocessor; by default all files are preprocessed if preprocessor is used
        print_n : {None}
          function for printing normal message
        print_w : {None}
          function for printing emphized warning message
        jobs : {1}
          concurrent compiling jobs
        """

        if print_n is None:
            self.print_n = print_fake
        else:
            self.print_n = print_n

        if print_w is None:
            self.print_w = print_fake
        else:
            self.print_w = print_w

        self.compiler = Compiler(cliargs=cliargs, print_w=self.print_w)

        self.jobs = cliargs.jobs
        self._sanitize_dirs(
            src_dir=cliargs.src,
            build_dir=cliargs.build_dir,
            obj_dir=cliargs.obj_dir,
            mod_dir=cliargs.mod_dir,
            lib_dir=cliargs.lib_dir,
            dinc=cliargs.include,
        )
        self._sanitize_files(
            libs=cliargs.libs, vlibs=cliargs.vlibs, ext_libs=cliargs.ext_libs, ext_vlibs=cliargs.ext_vlibs
        )
        self._set_preprocessor(
            preprocessor=cliargs.preprocessor,
            preprocessor_dir=cliargs.preprocessor_dir,
            preprocessor_ext=cliargs.preprocessor_ext,
            preprocessor_no_o=cliargs.preprocessor_no_o,
            preprocessor_args=cliargs.preprocessor_args,
        )

        self.ar = cliargs.ar
        self.arflags = cliargs.arflags
        self.ranlib = cliargs.ranlib

        self.cmd_comp = self.compiler.compile_cmd(mod_dir=self.mod_dir)
        self.cmd_link = self.compiler.link_cmd(mod_dir=self.mod_dir)
        return

    @staticmethod
    def get_fc(cliargs: Any) -> str:
        """
        Method for getting the compiler command built accordingly to the cli arguments.

        Parameters
        ----------
        cliargs : argparse object
        """
        return Compiler(cliargs=cliargs).fcs

    @staticmethod
    def get_cflags(cliargs: Any) -> str:
        """
        Method for getting the compiling flags built accordingly to the cli arguments.

        Parameters
        ----------
        cliargs : argparse object
        """
        return Compiler(cliargs=cliargs).cflags

    @staticmethod
    def get_lflags(cliargs: Any) -> str:
        """
        Method for getting the linking flags built accordingly to the cli arguments.

        Parameters
        ----------
        cliargs : argparse object
        """
        return Compiler(cliargs=cliargs).lflags

    @staticmethod
    def get_modsw(cliargs: Any) -> str:
        """
        Method for getting the compiler modules switch built accordingly to the cli arguments.

        Parameters
        ----------
        cliargs : argparse object
        """
        # modsw = cliargs.modsw
        # if cliargs.compiler.lower() in Builder.s_compilers and modsw == '':
        #   modsw = Builder.s_compilers[cliargs.compiler.lower()][2]
        # return modsw
        return Compiler(compiler=cliargs.compiler, modsw=cliargs.modsw).modsw

    def _sanitize_dirs(self, src_dir, build_dir, obj_dir, mod_dir, lib_dir, dinc):
        """
        Method for sanitizing directory paths.

        Parameters
        ----------
        src_dir : list
          directory into which search source files
        build_dir : str
          directory containing built files
        obj_dir : str
          directory containing compiled object files
        mod_dir : str
          directory containing .mod files
        lib_dir : list
          list of directories searched for libraries
        dinc : list
          list of directories for searching included files
        """
        self.src_dir = " ".join(src_dir)

        self.build_dir = os.path.normpath(build_dir)
        safe_mkdir(directory=self.build_dir, print_w=self.print_w)

        self.obj_dir = os.path.normpath(os.path.join(build_dir, obj_dir))
        safe_mkdir(directory=self.obj_dir, print_w=self.print_w)

        self.mod_dir = os.path.normpath(os.path.join(build_dir, mod_dir))
        safe_mkdir(directory=self.mod_dir, print_w=self.print_w)

        if lib_dir is None:
            self.lib_dir = []
        else:
            self.lib_dir = lib_dir
            for directory in self.lib_dir:
                directory = os.path.normpath(directory)
        if dinc is None:
            self.dinc = []
        else:
            self.dinc = dinc
            for directory in self.dinc:
                directory = os.path.normpath(directory)
        return

    def _sanitize_files(self, libs, vlibs, ext_libs, ext_vlibs):
        """
        Method for sanitizing files paths.

        Parameters
        ----------
        libs : list
          list of external libraries that are not into the path: must be passed with full paths
        vlibs : list
          list of external libraries that are not into the path and that are volatile (can changed thus triggering re-building): must be passed with full paths
        ext_libs : list
          list of external libraries that are into the path: are linked as '-llibrary.[a|so]' and searched into '-Ldir'
        ext_vlibs : list
          list of external libraries that are into the path and that are volatile (can changed thus triggering re-building): are linked as '-llibrary.[a|so]' and searched into '-Ldir'
        """
        if libs is None:
            self.libs = []
        else:
            self.libs = libs
            for lib in self.libs:
                lib = os.path.normpath(lib)
        if vlibs is None:
            self.vlibs = []
        else:
            self.vlibs = vlibs
            for lib in self.vlibs:
                lib = os.path.normpath(lib)
        if ext_libs is None:
            self.ext_libs = []
        else:
            self.ext_libs = ext_libs
            for lib in self.ext_libs:
                lib = os.path.normpath(lib)
        if ext_vlibs is None:
            self.ext_vlibs = []
        else:
            self.ext_vlibs = ext_vlibs
            for lib in self.ext_vlibs:
                lib = os.path.normpath(lib)
        return

    def _set_preprocessor(self, preprocessor, preprocessor_dir, preprocessor_ext, preprocessor_no_o, preprocessor_args):
        """
        Set preprocessor data.

        Parameters
        ----------
        preprocessor : bool
          activate preprocessor
        preprocessor_dir : str
          directory containing sources processed by preprocessor; by default are removed after used
        preprocessor_ext : list
          list of file extensions to be processed by preprocessor; by default all files are preprocessed if preprocessor is used
        """
        self.preprocessor = preprocessor
        if preprocessor_ext is None:
            self.preprocessor_ext = []
        else:
            self.preprocessor_ext = preprocessor_ext
        self.preprocessor_dir = preprocessor_dir
        if self.preprocessor_dir:
            self.preprocessor_dir = os.path.normpath(os.path.join(self.build_dir, self.preprocessor_dir))
            if not os.path.exists(self.preprocessor_dir):
                os.makedirs(self.preprocessor_dir)
        self.preprocessor_no_o = preprocessor_no_o
        self.preprocessor_args = preprocessor_args
        if self.preprocessor:
            preprocessor_exist = False
            for path in os.environ["PATH"].split(os.pathsep):
                preprocessor_exist = os.path.exists(os.path.join(path, self.preprocessor))
                if preprocessor_exist:
                    break
            if not preprocessor_exist:
                self.print_w(
                    "Error: " + self.preprocessor + " is not in your PATH! You cannot use -preprocessor switch."
                )
        return

    def _compile_preprocessor(self, file_to_compile):
        """
        Create compile command with preprocessor usage.

        Parameters
        ----------
        file_to_compile : ParsedFile object
          file to be compiled

        Returns
        -------
        str
          string containing the preprocessor command (empty if no preprocessing)
        str
          string containing the output file name of preprocessor
        bool
          True if the preprocessor output file should be removed after compilation
        """
        preprocessor_cmd = ""
        preprocessor_output = ""
        preprocessor_remove = False
        to_preprocess = False
        if self.preprocessor:
            if len(self.preprocessor_ext) > 0:
                if file_to_compile.extension in self.preprocessor_ext:
                    to_preprocess = True
            else:
                to_preprocess = True
            if to_preprocess and self.preprocessor_dir:
                preprocessor_dir = self.preprocessor_dir
                preprocessor_store = True
            else:
                preprocessor_dir = ""
                preprocessor_store = False
            if to_preprocess:
                output_prepend = " -o "
                if self.preprocessor_no_o:
                    output_prepend = " "
                preprocessor_cmd = (
                    self.preprocessor
                    + " "
                    + self.preprocessor_args
                    + " "
                    + file_to_compile.name
                    + output_prepend
                    + os.path.join(preprocessor_dir, file_to_compile.basename + ".pp.f90")
                )
                preprocessor_output = os.path.join(preprocessor_dir, file_to_compile.basename + ".pp.f90")
                if not preprocessor_store:
                    preprocessor_remove = True
        return preprocessor_cmd, preprocessor_output, preprocessor_remove

    def _compile_include(self):
        """
        Method for creating compile command with inluded paths.

        Returns
        -------
        str
          string containing the include command
        """
        include_cmd = ""
        if len(self.dinc) > 0:
            include_cmd = "".join(["-I" + s + " " for s in self.dinc])
        return include_cmd

    def _compile_command(self, file_to_compile):
        """
        Method for returning the OS commands for compiling file_to_compile.

        Parameters
        ----------
        file_to_compile : ParsedFile object
          file to be compiled

        Returns
        -------
        list
          list of (cmd_string, cleanup_path_or_None) tuples to execute in order;
          cleanup_path is set on the last step when a temporary preprocessor output must be removed
        """
        preprocessor_cmd, preprocessor_output, preprocessor_remove = self._compile_preprocessor(file_to_compile)
        include_cmd = self._compile_include()

        if preprocessor_cmd != "":
            compile_cmd = (
                self.cmd_comp
                + " "
                + include_cmd
                + preprocessor_output
                + " -o "
                + os.path.join(self.obj_dir, file_to_compile.basename + ".o")
            )
            cleanup = preprocessor_output if preprocessor_remove else None
            return [(preprocessor_cmd, None), (compile_cmd, cleanup)]
        else:
            comp_cmd = (
                self.cmd_comp
                + " "
                + include_cmd
                + file_to_compile.name
                + " -o "
                + os.path.join(self.obj_dir, file_to_compile.basename + ".o")
            )
            return [(comp_cmd, None)]

    def _link_command(self, file_to_build, output=None, nomodlibs=None, submodules=None, mklib=None):
        """
        Return the OS command for linkng file_to_build.

        Parameters
        ----------
        file_to_build : ParsedFile object
          file to be built
        output : str
          output name
        nomodlibs : {None}
          list of old-Fortran style libraries objects
        submodules : {None}
          list of submodules objects
        mklib : {None}
          activate building library mode

        Returns
        -------
        str
          string containing the link command
        """
        exe = self.get_output_name(file_to_build=file_to_build, output=output)
        link_cmd = (
            self.cmd_link
            + " "
            + self._get_libs_link_command(
                file_to_build=file_to_build, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib
            )
            + " -o "
            + exe
        )
        return link_cmd, exe

    def _mklib_command(self, file_to_build, output=None, nomodlibs=None, submodules=None, mklib=None):
        """
        Return the OS command for linkng file_to_build as a library.

        Parameters
        ----------
        file_to_build : ParsedFile object
          file to be built
        output : str
          output name
        nomodlibs : {None}
          list of old-Fortran style libraries objects
        submodules : {None}
          list of submodules objects
        mklib : {None}
          activate building library mode

        Returns
        -------
        str
          string containing the link command
        """
        lib = self.get_output_name(file_to_build=file_to_build, output=output, mklib=mklib)
        link_cmd = self._get_libs_link_command(
            file_to_build=file_to_build, exclude_programs=True, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib
        )
        ranlib_cmd = None
        if mklib is not None:
            if mklib.lower() == "shared":
                link_cmd = self.cmd_link + " " + link_cmd + " -o " + lib
            elif mklib.lower() == "static":
                link_cmd = self.ar + " " + self.arflags + " " + lib + " " + link_cmd
                if self.ranlib:
                    ranlib_cmd = self.ranlib + " " + lib
        return link_cmd, ranlib_cmd, lib

    def _get_libs_link_command(
        self, file_to_build, exclude_programs=False, nomodlibs=None, submodules=None, mklib=None
    ):
        """
        Return the libraries link command

        Parameters
        ----------
        file_to_build : ParsedFile object
          file to be built
        exclude_programs : {False}
          flag for excluding programs obj from the list
        nomodlibs : {None}
          list of old-Fortran style libraries objects
        submodules : {None}
          list of submodules objects
        mklib : {None}
          activate building library mode

        Returns
        -------
        str
          string containing the link command
        """
        objs = []
        if nomodlibs is not None:
            objs = objs + nomodlibs
        if submodules is not None:
            objs = objs + submodules
        objs = objs + file_to_build.obj_dependencies(exclude_programs=exclude_programs)
        objs = list(set(objs))  # remove duplicated objs
        link_cmd = (
            "".join([os.path.join(self.obj_dir, s + " ") for s in objs])
            + "".join([s + " " for s in self.libs])
            + "".join([s + " " for s in self.vlibs])
        )
        if mklib is None or mklib.lower() == "shared":
            if len(self.ext_libs) > 0:
                link_cmd += " " + "".join(["-l" + s + " " for s in self.ext_libs])
            if len(self.ext_vlibs) > 0:
                link_cmd += " " + "".join(["-l" + s + " " for s in self.ext_vlibs])
            if len(self.lib_dir) > 0:
                link_cmd += " " + "".join(["-L" + s + " " for s in self.lib_dir])
        return link_cmd

    def _get_hierarchy(self, file_to_build):
        """
        Create a hierarchy of compiling commands accordingly to the dependencies order.

        Parameters
        ----------
        file_to_build : ParsedFile
        """
        order_max = (
            max(
                [p for p in file_to_build.pfile_dep_all if not p.include and p.to_compile],
                key=operator.attrgetter("order"),
            ).order
            + 1
        )
        hierarchy = [[] for _ in range(order_max)]
        for dep in file_to_build.pfile_dep_all:
            if dep.to_compile and not dep.include:
                hierarchy[dep.order].append([dep.name, self._compile_command(file_to_compile=dep)])
                dep.to_compile = False
        hierarchy = [h for h in hierarchy if len(h) > 0]
        return hierarchy

    def _compile(self, file_to_build, verbose=False, log=False, quiet=False):
        """
        Compile file.

        Parameters
        ----------
        file_to_build : ParsedFile
        verbose : {False}
          bool for activate extreme verbose outputs
        log : {False}
          bool for activate errors log saving

        Returns
        -------
        build_ok : bool
          flag for checking the building is completed;
          it is used for purging out failed non-module-library in order to try to build the target anywhere;
        """
        build_ok = True
        hierarchy = self._get_hierarchy(file_to_build)
        for deps in reversed(hierarchy):
            files_to_compile = ""
            steps_per_file = []
            for dep in deps:
                files_to_compile = files_to_compile + " " + dep[0]
                steps_per_file.append(dep[1])
            if len(deps) > 1 and self.jobs > 1 and __parallel__:
                jobs = min(len(deps), self.jobs)
                if not quiet:
                    self.print_n("Compiling" + files_to_compile + " using " + str(jobs) + " concurrent processes")
                    if verbose:
                        self.print_n("Executing: " + str(steps_per_file))
                pool = Pool(processes=jobs)
                results = pool.map(syswork_steps, steps_per_file)
                pool.close()
                pool.join()
            else:
                if not quiet:
                    self.print_n("Compiling" + files_to_compile + " serially")
                results = []
                for steps in steps_per_file:
                    if not quiet:
                        if verbose:
                            self.print_n("Executing: " + str(steps))
                    result = syswork_steps(steps)
                    results.append(result)
            if not file_to_build.nomodlib:
                if log:
                    check_results(results=results, log="building-errors.log", print_w=self.print_w)
                else:
                    check_results(results=results, print_w=self.print_w)
        if file_to_build.nomodlib:
            if results[0][0] != 0:
                build_ok = False
                if not quiet:
                    self.print_w(
                        "Warning: compiling file "
                        + file_to_build.name
                        + " fails! Removing from the list of non-module libraries linked into the target!"
                    )
        return build_ok

    def gnu_make(self) -> str:
        """
        Return the builder options formated as GNU Make variables

        Returns
        -------
        str
          string containing the builder options in GNU Make format
        """
        string = []
        string.append("DOBJ    = " + os.path.normpath(self.obj_dir) + "/\n")
        string.append("DMOD    = " + os.path.normpath(self.mod_dir) + "/\n")
        string.append("DEXE    = " + os.path.normpath(self.build_dir) + "/\n")
        libs = []
        if len(self.libs) > 0:
            libs += self.libs
        if len(self.vlibs) > 0:
            libs += self.vlibs
        if len(self.ext_libs) > 0:
            libs += ["-l" + l + " " for l in self.ext_libs]
        if len(self.ext_vlibs) > 0:
            libs += ["-l" + l + " " for l in self.ext_vlibs]
        if len(self.lib_dir) > 0:
            libs += ["-L" + l + " " for l in self.lib_dir]
        if len(libs) > 0:
            string.append("LIBS    = " + "".join(" " + l for l in libs) + "\n")
        else:
            string.append("LIBS    =\n")
        string.append("FC      = " + self.compiler.fcs + "\n")
        string.append("OPTSC   = " + self.compiler.cflags + " " + self.compiler.modsw + self.mod_dir + "\n")
        string.append("OPTSL   = " + self.compiler.lflags + " " + self.compiler.modsw + self.mod_dir + "\n")
        return "".join(string)

    def build(
        self,
        file_to_build: Any,
        output: Optional[str] = None,
        nomodlibs: Optional[list[str]] = None,
        submodules: Optional[list[str]] = None,
        mklib: Optional[str] = None,
        verbose: bool = False,
        log: bool = False,
        quiet: bool = False,
        track: bool = False,
    ) -> bool:
        """
        Build a file.

        Parameters
        ----------
        file_to_build : ParsedFile
        output : str
          output name
        nomodlibs : {None}
          list of old-Fortran style libraries objects
        submodules : {None}
          list of submodules objects
        mklib : {None}
          bool for activate building library mode
        verbose : {False}
          bool for activate extreme verbose outputs
        log : {False}
          bool for activate errors log saving
        quiet : {False}
          bool for activate quiet mode
        track : {False}
          bool for activate build traking mode for subsequent install command

        Returns
        -------
        build_ok : bool
          flag for checking the building is completed;
          it is used for purging out failed non-module-library in order to try to build the target anywhere;
        """
        build_ok = True
        if not quiet:
            self.print_n("Building " + file_to_build.name)
        # sort dependencies accordingly to indirect dependency
        file_to_build.sort_dependencies()
        # creating a hierarchy list of compiling commands accordingly to the order of all dependencies
        if len([p for p in file_to_build.pfile_dep_all if not p.include and p.to_compile]) > 0:
            build_ok = self._compile(file_to_build=file_to_build, verbose=verbose, log=log, quiet=quiet)
        else:
            if not quiet:
                self.print_n("Nothing to compile, all objects are up-to-date")
        if file_to_build.program:
            link_cmd, exe = self._link_command(
                file_to_build=file_to_build, output=output, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib
            )
            if not quiet:
                self.print_n("Linking " + exe)
                if verbose:
                    self.print_n("Executing: " + str(link_cmd))
            result = syswork(link_cmd)
            if log:
                check_results(results=[result], log="building-errors.log", print_w=self.print_w)
            else:
                check_results(results=[result], print_w=self.print_w)
            if not quiet:
                self.print_n("Target " + file_to_build.name + " has been successfully built")
        elif mklib:
            link_cmd, ranlib_cmd, lib = self._mklib_command(
                file_to_build=file_to_build, output=output, nomodlibs=nomodlibs, submodules=submodules, mklib=mklib
            )
            if not quiet:
                self.print_n("Linking " + lib)
                if verbose:
                    self.print_n("Executing: " + str(link_cmd))
            result = syswork(link_cmd)
            if log:
                check_results(results=[result], log="building-errors.log", print_w=self.print_w)
            else:
                check_results(results=[result], print_w=self.print_w)
            if ranlib_cmd and result[0] == 0:
                result = syswork(ranlib_cmd)
                if log:
                    check_results(results=[result], log="building-errors.log", print_w=self.print_w)
                else:
                    check_results(results=[result], print_w=self.print_w)
            if not quiet:
                self.print_n("Target " + file_to_build.name + " has been successfully built")
        if track:
            print("Track building of " + file_to_build.name)
            track_file_name = self.get_track_build_file(file_to_build)
            with open(track_file_name, "w") as track_file:
                track_file.writelines("[build]\n")
                track_file.writelines(
                    "output = " + self.get_output_name(file_to_build=file_to_build, output=output) + "\n"
                )
                if file_to_build.program:
                    track_file.writelines("program = True\n")
                elif mklib:
                    track_file.writelines("library = True\n")
                    track_file.writelines(
                        "mod_file = " + os.path.join(self.mod_dir, file_to_build.basename + ".mod") + "\n"
                    )
        return build_ok

    def get_output_name(self, file_to_build: Any, output: Optional[str] = None, mklib: Optional[str] = None) -> str:
        """
        Return the output build file name.

        Parameters
        ----------
        file_to_build : ParsedFile
        output : str
          output build file name
        mklib : {None}
          activate building library mode

        Returns
        -------
        output : str
          output build file name
        """
        if output:
            build_name = os.path.join(self.build_dir, output)
        else:
            if mklib is not None:
                if mklib.lower() == "shared":
                    build_name = os.path.join(self.build_dir, file_to_build.basename + ".so")
                elif mklib.lower() == "static":
                    build_name = os.path.join(self.build_dir, file_to_build.basename + ".a")
            else:
                build_name = os.path.join(self.build_dir, file_to_build.basename)
        return build_name

    def get_track_build_file(self, file_to_build: Any) -> str:
        """
        Return the file name of the 'track build' file.

        Parameters
        ----------
        file_to_build : ParsedFile

        Returns
        -------
        track_file_name : str
          file name of 'track build' file
        """
        return os.path.join(self.build_dir, "." + os.path.basename(file_to_build.name) + ".track_build")

    def verbose(self, quiet: bool = False) -> str:
        """
        The method verbose returns a verbose message containing builder infos.

        Parameters
        ----------
        quiet : {False}
          flag for making less verbose outputs
        """
        message = ""
        if not quiet:
            message = "Builder options" + "\n"
            message += "  Directories\n"
            message += '    Root of source files directory: "' + self.src_dir + '"\n'
            message += '    Building directory: "' + self.build_dir + '"\n'
            message += '    Compiled-objects .o   directory: "' + self.obj_dir + '"\n'
            message += '    Compiled-objects .mod directory: "' + self.mod_dir + '"\n'
            if len(self.lib_dir) > 0:
                message += "  External libraries directories: " + "".join([s + " " for s in self.lib_dir]) + "\n"
            if len(self.dinc) > 0:
                message += "  Included paths: " + "".join([s + " " for s in self.dinc]) + "\n"
            if len(self.libs) > 0:
                message += "  Linked libraries with full path: " + "".join([s + " " for s in self.libs]) + "\n"
            if len(self.vlibs) > 0:
                message += (
                    "  Linked volatile libraries with full path: " + "".join([s + " " for s in self.vlibs]) + "\n"
                )
            if len(self.ext_libs) > 0:
                message += "  Linked libraries in path: " + "".join([s + " " for s in self.ext_libs]) + "\n"
            if len(self.ext_vlibs) > 0:
                message += "  Linked volatile libraries in path: " + "".join([s + " " for s in self.ext_vlibs]) + "\n"
            message += self.compiler.pprint(prefix="  ")
            message += "  Archiver: " + self.ar + " " + self.arflags + "\n"
            if self.ranlib:
                message += "  Ranlib: " + self.ranlib + "\n"
            message += "  Preprocessor used: " + str(self.preprocessor) + "\n"
            message += "  Preprocessor args: " + str(self.preprocessor_args) + "\n"
            message += "  Preprocessor output directory: " + str(self.preprocessor_dir) + "\n"
            message += "  Preprocessor extensions processed: " + str(self.preprocessor_ext) + "\n"
        return message



#!/usr/bin/env python
# File: src/main/python/fobis/Compiler.py
# Author: Stefano Zaghi <stefano.zaghi@gmail.com>
# Date: 01.08.2017
# Last Modified Date: 01.08.2017
# Last Modified By: Stefano Zaghi <stefano.zaghi@gmail.com>
"""
Compiler.py, module definition of Compiler class.
This is a class designed for handling compilers default support.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from builtins import *
import re

from .utils import print_fake

__regex_opts__ = re.compile(r"-O[0-9,s]")


class Compiler:
    """
    Compiler is an object that handles the compilers default support, its attributes and methods.

    Attributes
    ----------
    supported : {['gnu', 'intel', 'intel_nextgen', 'g95', 'opencoarrays-gnu', 'pgi', 'ibm', 'nag', 'nvfortran', 'amd', 'custom']}
      list of supported compilers
    """

    supported = [
        "gnu",
        "intel",
        "intel_nextgen",
        "g95",
        "opencoarrays-gnu",
        "pgi",
        "ibm",
        "nag",
        "nvfortran",
        "amd",
        "custom",
    ]

    def __init__(self, cliargs, print_w=None):
        """
        Parameters
        ----------
        cliargs : argparse object
        print_w : {None}
          function for printing emphized warning message

        Attributes
        ----------
        compiler : {None}
          str containing compiler vendor name
        fcs : {None}
          str containing compiler statement
        cflags : {None}
          str containing compiling flags
        lflags : {None}
          str containing linking flags
        preproc : {None}
          str containing preprocessing flags
        modsw : {None}
          str containing compiler switch for modules searching path
        mpi : {False}
          activate the MPI compiler
        openmp : {False}
          activate the OpenMP pragmas
        openmp_offload : {False}
          activate the OpenMP Offload pragmas
        coarray : {False}
          activate the coarray compilation
        coverage : {False}
          activate the coverage instruments
        profile : {False}
          activate the profile instruments
        print_w : {None}
          function for printing emphized warning message
        """
        if print_w is None:
            self.print_w = print_fake
        else:
            self.print_w = print_w

        self._mpi = None
        self._openmp = None
        self._openmp_offload = None
        self._coarray = None
        self._coverage = None
        self._profile = None
        self.compiler = cliargs.compiler
        if self.compiler:
            if self.compiler.lower() == "gnu":
                self._gnu()
            elif self.compiler.lower() == "intel":
                self._intel()
            elif self.compiler.lower() == "intel_nextgen":
                self._intel_nextgen()
            elif self.compiler.lower() == "g95":
                self._g95()
            elif self.compiler.lower() == "opencoarrays-gnu":
                self._opencoarrays_gnu()
            elif self.compiler.lower() == "pgi":
                self._pgi()
            elif self.compiler.lower() == "ibm":
                self._ibm()
            elif self.compiler.lower() == "nag":
                self._nag()
            elif self.compiler.lower() == "nvfortran":
                self._nvfortran()
            elif self.compiler.lower() == "amd":
                self._amd()
            elif self.compiler.lower() == "custom":
                self._custom()
            else:
                self._gnu()
        # overriding default values if passed
        if cliargs.fc:
            self.fcs = cliargs.fc
        if cliargs.cflags:
            self.cflags = cliargs.cflags
        if cliargs.lflags:
            self.lflags = cliargs.lflags
        if cliargs.preproc:
            self.preproc = cliargs.preproc
        if cliargs.modsw:
            self.modsw = cliargs.modsw.replace("'", "")
        self.mpi = cliargs.mpi
        self.openmp = cliargs.openmp
        self.openmp_offload = cliargs.openmp_offload
        self.coarray = cliargs.coarray
        self.coverage = cliargs.coverage
        self.profile = cliargs.profile
        self._set_fcs()
        self._set_cflags()
        self._set_lflags(mklib=cliargs.mklib)
        return

    def __str__(self):
        return self.pprint()

    def _gnu(self):
        """Set compiler defaults to the GNU gfortran compiler options."""
        self.compiler = "gnu"
        self.fcs = "gfortran"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-J "
        self._mpi = "mpif90"
        self._openmp = ["-fopenmp", "-fopenmp"]
        self._coarray = ["-fcoarray=lib", "-fcoarray=lib -lcaf_mpi"]
        self._coverage = ["-ftest-coverage -fprofile-arcs", "-fprofile-arcs"]
        self._profile = ["-pg", "-pg"]
        return

    def _intel(self):
        """Set compiler defaults to the Intel Fortran compiler options."""
        self.compiler = "intel"
        self.fcs = "ifort"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-module "
        self._mpi = "mpiifort"
        self._openmp = ["-qopenmp", "-qopenmp"]
        self._coarray = ["-coarray", "-coarray"]
        self._coverage = ["-prof-gen=srcpos", ""]
        self._profile = ["", ""]
        return

    def _intel_nextgen(self):
        """Set compiler defaults to the Intel Next Generation Fortran compiler options."""
        self.compiler = "intel_nextgen"
        self.fcs = "ifx"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-module "
        self._mpi = "mpiifort -fc=ifx"
        self._openmp = ["-qopenmp", "-fiopenmp"]
        self._openmp_offload = "-fopenmp-targets=spir64"
        self._coarray = ["-coarray", "-coarray"]
        self._coverage = ["-prof-gen=srcpos", ""]
        self._profile = ["", ""]
        return

    def _g95(self):
        """Set compiler defaults to the g95 compiler options."""
        self.compiler = "g95"
        self.fcs = "g95"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-fmod="
        self._mpi = "mpif90"
        self._openmp = ["", ""]
        self._coarray = ["", ""]
        self._coverage = ["", ""]
        self._profile = ["", ""]
        return

    def _opencoarrays_gnu(self):
        """Set compiler defaults to the OpenCoarrays-Aware GNU gfortran compiler options."""
        self.compiler = "opencoarrays-gnu"
        self.fcs = "caf"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-J "
        self._mpi = "mpif90"
        self._openmp = ["-fopenmp", "-fopenmp"]
        self._coarray = ["", ""]
        self._coverage = ["-ftest-coverage -fprofile-arcs", "-fprofile-arcs"]
        self._profile = ["-pg", "-pg"]
        return

    def _pgi(self):
        """Set compiler defaults to the PGI compiler options."""
        self.compiler = "pgi"
        self.fcs = "pgfortran"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-module "
        self._mpi = "mpif90"
        self._openmp = ["-mp", "-mp"]
        self._coarray = ["", ""]
        self._coverage = ["", ""]
        self._profile = ["-pg", "-pg"]
        return

    def _ibm(self):
        """Set compiler defaults to the IBM XL Fortran compiler options."""
        self.compiler = "ibm"
        self.fcs = "xlf2008_r"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-qmoddir="
        self._mpi = "mpif90"
        self._openmp = ["-qsmp=omp", "-qsmp=omp"]
        self._coarray = ["", ""]
        self._coverage = ["", ""]
        self._profile = ["-pg", "-pg"]
        return

    def _nag(self):
        """Set compiler defaults to the NAG fortran compiler options."""
        self.compiler = "gnu"
        self.fcs = "nagfor"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-mdir "
        self._mpi = "mpif90"
        self._openmp = ["-openmp", "-openmp"]
        self._coarray = ["-num_images=1", ""]
        self._coverage = ["", ""]
        self._profile = ["-pg", "-pg"]
        return

    def _nvfortran(self):
        """Set compiler defaults to the NVFortran compiler options."""
        self.compiler = "nvfortran"
        self.fcs = "nvfortran"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-module "
        self._mpi = "mpif90"
        self._openmp = ["-mp", "-mp"]
        self._coarray = ["", ""]
        self._coverage = ["", ""]
        self._profile = ["-pg", "-pg"]
        return

    def _amd(self):
        """Set compiler defaults to the AMD Flang compiler options."""
        self.compiler = "amd"
        self.fcs = "amdflang"
        self.cflags = "-c"
        self.lflags = ""
        self.preproc = ""
        self.modsw = "-module-dir "
        self._mpi = "mpif90"
        self._openmp = ["-fopenmp", "-fopenmp"]
        self._coarray = ["", ""]
        self._coverage = ["", ""]
        self._profile = ["-pg", "-pg"]
        return

    def _custom(self):
        """Set compiler defaults to be empty."""
        self.compiler = ""
        self.fcs = ""
        self.cflags = ""
        self.lflags = ""
        self.preproc = ""
        self.modsw = ""
        self._mpi = ""
        self._openmp = ""
        self._coarray = ""
        self._coverage = ""
        self._profile = ""
        return

    def _set_fcs(self):
        """Set the compiler command statement directly depending on the compiler."""
        if self.compiler.lower() in Compiler.supported:
            if self.mpi:
                self.fcs = self._mpi
        return

    def _set_cflags(self):
        """Set the compiling flags directly depending on the compiler."""
        if self.coverage:
            if self._coverage[0] != "":
                if re.search(__regex_opts__, self.cflags):
                    self.print_w(
                        "Warning: found optimizations cflags within coverage ones: coverage results can be alterated!"
                    )
                self.cflags += " -Og " + self._coverage[0]
        if self.profile:
            if self._profile[0] != "":
                self.cflags += " " + self._profile[0]
        if self.openmp:
            if self._openmp[0] != "":
                self.cflags += " " + self._openmp[0]
        if self.openmp_offload:
            if self._openmp_offload != "":
                self.cflags += " " + self._openmp_offload
        if self.coarray:
            if self._coarray[0] != "":
                self.cflags += " " + self._coarray[0]
        if self.preproc is not None:
            if self.preproc != "":
                self.cflags += " " + self.preproc
        self.cflags = re.sub(r" +", r" ", self.cflags)
        return

    def _set_lflags(self, mklib=None):
        """Set the linking flags directly depending on the compiler."""
        if mklib is not None:
            if mklib.lower() == "shared":
                if "-shared" not in self.lflags:
                    self.lflags += " -shared"
        if self.coverage:
            if self._coverage[1] != "":
                if re.search(__regex_opts__, self.lflags):
                    self.print_w(
                        "Warning: found optimizations lflags within coverage ones: coverage results can be alterated!"
                    )
                self.lflags += " -Og " + self._coverage[1]
        if self.profile:
            if self._profile[1] != "":
                self.lflags += " " + self._profile[1]
        if self.openmp:
            if self._openmp[1] != "":
                self.lflags += " " + self._openmp[1]
        if self.openmp_offload:
            if self._openmp_offload != "":
                self.lflags += " " + self._openmp_offload
        if self.coarray:
            if self._coarray[1] != "":
                self.lflags += " " + self._coarray[1]
        self.lflags = re.sub(r" +", r" ", self.lflags)
        return

    def compile_cmd(self, mod_dir):
        """
        Method returning the compile command accordingly to the compiler options.

        Parameters
        ----------
        mod_dir : str
          path of the modules directory
        """
        return self.fcs + " " + self.cflags + " " + self.modsw + mod_dir + " -I" + mod_dir

    def link_cmd(self, mod_dir):
        """
        Method returning the compile command accordingly to the compiler options.

        Parameters
        ----------
        mod_dir : str
          path of the modules directory
        """
        return self.fcs + " " + self.lflags + " " + self.modsw + mod_dir

    def pprint(self, prefix=""):
        """
        Pretty printer.

        Parameters
        ----------
        prefix : {''}
          prefixing string of each line
        """
        string = prefix + "Compiler options\n"
        string += prefix + '  Vendor: "' + self.compiler.strip() + '"\n'
        string += prefix + '  Compiler command: "' + self.fcs.strip() + '"\n'
        string += prefix + '  Module directory switch: "' + self.modsw.strip() + '"\n'
        string += prefix + '  Compiling flags: "' + self.cflags.strip() + '"\n'
        string += prefix + '  Linking flags: "' + self.lflags.strip() + '"\n'
        string += prefix + '  Preprocessing flags: "' + self.preproc.strip() + '"\n'
        string += prefix + "  Coverage: " + str(self.coverage) + "\n"
        if self.coverage:
            string += prefix + "    Coverage compile and link flags: " + str(self._coverage) + "\n"
        string += prefix + "  Profile: " + str(self.profile) + "\n"
        if self.profile:
            string += prefix + "    Profile compile and link flags: " + str(self._profile) + "\n"
        return string



#!/usr/bin/env python
# File: /home/stefano/python/FoBiS/src/main/python/fobis/Fobos.py
# Author: Stefano Zaghi <stefano.zaghi@gmail.com>
# Date: 28.08.2017
# Last Modified Date: 28.08.2017
# Last Modified By: Stefano Zaghi <stefano.zaghi@gmail.com>
"""
fobos.py, module definition of fobos class.
This is a class aimed at fobos file handling.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
import configparser
import os
import re
import sys
from collections.abc import Callable
from copy import deepcopy
from typing import Any, Optional

from .utils import check_results, print_fake, syswork


class Fobos:
    """
    Fobos is an object that handles fobos file, its attributes and methods.
    """

    def __init__(
        self,
        cliargs: Any,
        print_n: Optional[Callable[..., None]] = None,
        print_w: Optional[Callable[..., None]] = None,
    ) -> None:
        """
        Parameters
        ----------
        cliargs : argparse object
        print_n : {None}
          function for printing normal message
        print_w : {None}
          function for printing emphized warning message
        """
        if print_n is None:
            self.print_n = print_fake
        else:
            self.print_n = print_n

        if print_w is None:
            self.print_w = print_fake
        else:
            self.print_w = print_w

        self.fobos = None
        self.mode = None
        self.local_variables = {}
        if cliargs.fobos:
            filename = cliargs.fobos
        else:
            filename = "fobos"
        if os.path.exists(filename):
            self.fobos = configparser.RawConfigParser()
            if not cliargs.fobos_case_insensitive:
                self.fobos.optionxform = str  # case sensitive
            self.fobos.read(filename)
            self._set_cliargs(cliargs=cliargs)
        return

    def _check_mode(self, mode):
        """
        Function for checking the presence of the selected mode into the set defined inside the fobos.

        Parameters
        ----------
        mode : str
          name of the selcted mode
        """
        if self.fobos:
            if self.fobos.has_option("modes", "modes"):
                if mode in self.fobos.get("modes", "modes"):
                    self.mode = mode
                else:
                    self.print_w('Error: the mode "' + mode + '" is not defined into the fobos file.')
                    self.modes_list()
                    sys.exit(1)
            else:
                self.print_w('Error: fobos file has not "modes" section.')
                sys.exit(1)
        return

    def _set_mode(self, mode=None):
        """
        Function for setting the selected mode.

        Parameters
        ----------
        mode : {None}
          selected mode
        """
        if self.fobos:
            if mode:
                self._check_mode(mode=mode)
            else:
                if self.fobos.has_option("modes", "modes"):
                    self.mode = self.fobos.get("modes", "modes").split()[0]  # first mode selected
                else:
                    if self.fobos.has_section("default"):
                        self.mode = "default"
                    else:
                        self.print_w('Warning: fobos file has not "modes" section neither "default" one')
        return

    def _check_template(self):
        """
        Check and apply template sections.

        Each mode may specify one or more template names as a space-separated
        list.  Templates are applied left-to-right: the mode itself always
        wins, and earlier templates take precedence over later ones.
        Template-of-template inheritance is expanded depth-first.
        Circular references are detected and cause an error.
        """
        if not self.fobos:
            return

        def _template_names(section):
            if self.fobos.has_option(section, "template"):
                return self.fobos.get(section, "template").split()
            return []

        def _resolve(section, visiting):
            """Return ordered list of template names to apply for section."""
            if section in visiting:
                self.print_w('Error: circular template reference detected involving "' + section + '"')
                sys.exit(1)
            visiting = visiting | {section}
            resolved = []
            for name in _template_names(section):
                if not self.fobos.has_section(name):
                    self.print_w('Error: mode "' + section + '" uses template "' + name + '" that is NOT defined')
                    sys.exit(1)
                if name not in resolved:
                    resolved.append(name)
                for sub in _resolve(name, visiting):
                    if sub not in resolved:
                        resolved.append(sub)
            return resolved

        for section in self.fobos.sections():
            if not self.fobos.has_option(section, "template"):
                continue
            for template in _resolve(section, set()):
                for item in self.fobos.items(template):
                    if item[0] == "template":
                        continue
                    if not self.fobos.has_option(section, item[0]):
                        self.fobos.set(section, item[0], item[1])
        return

    def _get_local_variables(self):
        """
        Get the definition of local variables defined into any sections (modes).
        """
        if self.fobos:
            for section in self.fobos.sections():
                for item in self.fobos.items(section):
                    if item[0].startswith("$"):
                        self.local_variables[item[0]] = item[1].replace("\n", " ")
        return

    def _substitute_local_variables_mode(self):
        """
        Substitute the definition of local variables defined into the mode (section) selected.
        """
        if self.fobos and self.mode:
            self._substitute_local_variables_section(section=self.mode)
        return

    def _substitute_local_variables_section(self, section):
        """
        Substitute the definition of local variables defined into a section.
        """
        if self.fobos:
            if self.fobos.has_section(section):
                for item in self.fobos.items(section):
                    item_val = item[1]
                    for key, value in list(self.local_variables.items()):
                        item_val = re.sub(re.escape(key), value, item_val)
                        # item_val = re.sub(r"(?!" + re.escape(key) + r"[aZ_-])\s*" + re.escape(key) + r"\s*", value, item_val)
                    self.fobos.set(section, item[0], item_val)
        return

    def _check_local_variables(self):
        """
        Get and substitute the definition of local variables defined into any sections (modes).
        """
        if self.fobos:
            self._get_local_variables()
            if len(self.local_variables) > 0:
                self._substitute_local_variables_mode()
        return

    def _set_cliargs_attributes(self, cliargs, cliargs_dict):
        """
        Set attributes of cliargs from fobos options.

        Parameters
        ----------
        cliargs : argparse object
        cliargs_dict : argparse object attributes dictionary
        """
        if self.mode:
            for item in self.fobos.items(self.mode):
                if item[0] in cliargs_dict:
                    if isinstance(cliargs_dict[item[0]], bool):
                        setattr(cliargs, item[0], self.fobos.getboolean(self.mode, item[0]))
                    elif isinstance(cliargs_dict[item[0]], int):
                        setattr(cliargs, item[0], int(item[1]))
                    elif isinstance(cliargs_dict[item[0]], list):
                        setattr(cliargs, item[0], item[1].split())
                    else:
                        setattr(cliargs, item[0], item[1])
        return

    @staticmethod
    def _check_cliargs_cflags(cliargs, cliargs_dict):
        """
        Method for setting attribute of cliargs.

        Parameters
        ----------
        cliargs : argparse object
        cliargs_dict : argparse object attributes dictionary
        """
        for item in cliargs_dict:
            if item in ["cflags", "lflags", "preproc"]:
                val_cli = cliargs_dict[item]
                val_fobos = getattr(cliargs, item)
                if item == "cflags":
                    if val_cli == "-c":
                        match = re.search(r"(-c\s+|-c$)", val_fobos)
                        if match:
                            val_cli = ""  # avoid multiple -c flags
                if val_fobos and val_cli:
                    setattr(cliargs, item, val_fobos + " " + val_cli)
        return

    def _set_cliargs(self, cliargs):
        """
        Set cliargs from fobos options.

        Parameters
        ----------
        cliargs : argparse object
        """
        if self.fobos:
            cliargs_dict = deepcopy(cliargs.__dict__)
            self._set_mode(mode=cliargs.mode)
            self._check_template()
            self._check_local_variables()
            self._set_cliargs_attributes(cliargs=cliargs, cliargs_dict=cliargs_dict)
            self._check_cliargs_cflags(cliargs=cliargs, cliargs_dict=cliargs_dict)
        return

    def get(self, option: str, mode: Optional[str] = None, toprint: bool = True) -> Optional[str]:
        """
        Get options defined into the fobos file.

        Parameters
        ----------
        option : str
          option name
        mode : {None}
          eventual mode name
        toprint : {True}
          return of the value: if toprint==False the value is return otherwise is printed to stdout
        """
        value = ""
        if self.fobos:
            self._set_mode(mode=mode)
            if self.fobos.has_option(self.mode, option):
                value = self.fobos.get(self.mode, option)
        if toprint:
            # self.print_w(value)
            print(value)
            return
        else:
            return value

    def get_output_name(self, mode: Optional[str] = None, toprint: bool = True) -> Optional[str]:
        """
        Method for building the output name accordingly to the fobos options.

        Parameters
        ----------
        mode : {None}
          eventual mode name
        toprint : {True}
          return of the value: if toprint==False the value is return otherwise is printed to stdout
        """
        output = ""
        build_dir = self.get(option="build_dir", mode=mode, toprint=False)
        mklib = self.get(option="mklib", mode=mode, toprint=False)
        if self.fobos:
            self._set_mode(mode=mode)
            if self.fobos.has_option(self.mode, "output"):
                output = self.fobos.get(self.mode, "output")
                output = os.path.normpath(os.path.join(build_dir, output))
            elif self.fobos.has_option(self.mode, "target"):
                output = self.fobos.get(self.mode, "target")
                output = os.path.splitext(os.path.basename(output))[0]
                if mklib.lower() == "shared":
                    output = output + ".so"
                elif mklib.lower() == "static":
                    output = output + ".a"
                output = os.path.normpath(os.path.join(build_dir, output))
        if toprint:
            # self.print_w(output)
            print(output)
            return
        else:
            return output

    def modes_list(self) -> None:
        """List defined modes."""
        if self.fobos:
            self.print_n("The fobos file defines the following modes:")
            if self.fobos.has_option("modes", "modes"):
                modes = self.fobos.get("modes", "modes").split()
                for mode in modes:
                    if self.fobos.has_section(mode):
                        if self.fobos.has_option(mode, "help"):
                            helpmsg = self.fobos.get(mode, "help")
                        else:
                            helpmsg = ""
                        self.print_n('  - "' + mode + '" ' + helpmsg)
            else:
                self.print_w("Error: no modes are defined into the fobos file!")
                sys.exit(1)
        sys.exit(0)
        return

    @staticmethod
    def print_template(cliargs: Any) -> None:
        """
        Print fobos template.

        Parameters
        ----------
        cliargs : argparse object
        """
        print("[default]")
        for argument in vars(cliargs):
            attribute = getattr(cliargs, argument)
            if isinstance(attribute, list):
                attribute = " ".join(attribute)
            print(str(argument) + " = " + str(attribute))

    def get_project_info(self) -> dict[str, Any]:
        """
        Parse [project] section and return project metadata.

        Returns
        -------
        dict
          dict with keys 'name' (str), 'authors' (list of str),
          'version' (str, raw value as written in fobos — not resolved),
          'summary' (str), 'repository' (str), 'website' (str), and 'email' (str).
          All values are empty/empty-list if the section or option is absent.
        """
        info = {"name": "", "authors": [], "version": "", "summary": "", "repository": "", "website": "", "email": ""}
        if self.fobos and self.fobos.has_section("project"):
            if self.fobos.has_option("project", "name"):
                info["name"] = self.fobos.get("project", "name").strip()
            if self.fobos.has_option("project", "authors"):
                raw = self.fobos.get("project", "authors")
                info["authors"] = [a.strip() for a in raw.splitlines() if a.strip()]
            if self.fobos.has_option("project", "version"):
                info["version"] = self.fobos.get("project", "version").strip()
            if self.fobos.has_option("project", "summary"):
                info["summary"] = self.fobos.get("project", "summary").strip()
            if self.fobos.has_option("project", "repository"):
                info["repository"] = self.fobos.get("project", "repository").strip()
            if self.fobos.has_option("project", "website"):
                info["website"] = self.fobos.get("project", "website").strip()
            if self.fobos.has_option("project", "email"):
                info["email"] = self.fobos.get("project", "email").strip()
        return info

    def get_version(self) -> str:
        """
        Resolve the project version from [project] and/or git tags.

        Resolution steps
        ----------------
        1. Read 'version' from [project] in fobos.  If the value is a
           file path (relative to the git repository root), the version
           string is read from that file.
        2. Query the most recent git tag via ``git describe --tags --abbrev=0``.
        3. If both sources provide a version and they disagree, emit a
           warning with a suggested fix.
        4. Return the fobos version when available; fall back to the git
           tag; return '' when neither source is determinable.

        Returns
        -------
        str
          Resolved version string, or '' if not determinable.
        """
        fobos_version = ""
        if self.fobos and self.fobos.has_section("project"):
            if self.fobos.has_option("project", "version"):
                raw = self.fobos.get("project", "version").strip()
                # try to resolve as a file path relative to the git repo root
                git_root_result = syswork("git rev-parse --show-toplevel")
                if git_root_result[0] == 0:
                    candidate = os.path.join(git_root_result[1].strip(), raw)
                    if os.path.isfile(candidate):
                        with open(candidate) as ver_file:
                            fobos_version = ver_file.read().strip()
                    else:
                        fobos_version = raw
                else:
                    fobos_version = raw  # not inside a git repo; treat as literal
                if fobos_version and not fobos_version.startswith("v"):
                    fobos_version = "v" + fobos_version

        # query the most recent git tag
        git_version = ""
        git_result = syswork("git describe --tags --abbrev=0")
        if git_result[0] == 0:
            git_version = git_result[1].strip()

        # warn on mismatch
        if fobos_version and git_version and fobos_version != git_version:
            git_version_v = git_version if git_version.startswith("v") else "v" + git_version
            self.print_w("Warning: project version mismatch!")
            self.print_w("  fobos [project] version : " + fobos_version)
            self.print_w("  git tag version         : " + git_version)
            self.print_w("  To fix, either:")
            self.print_w("    - update fobos: set  version = " + git_version_v + "  under [project]")
            self.print_w("    - create a matching tag: git tag " + fobos_version + " && git push --tags")

        return fobos_version or git_version

    def get_dependencies(self) -> dict[str, str]:
        """
        Parse [dependencies] section and return dict of {name: spec_string}.

        Returns
        -------
        dict
          mapping of dependency name to its spec string, or empty dict if no section
        """
        deps = {}
        if self.fobos and self.fobos.has_section("dependencies"):
            for name, spec in self.fobos.items("dependencies"):
                if name == "deps_dir":
                    continue
                deps[name] = spec
        return deps

    def get_deps_dir(self, default: str = ".fobis_deps") -> str:
        """
        Read deps_dir from [dependencies] section of fobos.

        Parameters
        ----------
        default : str
          value returned when the option is absent [default: '.fobis_deps']

        Returns
        -------
        str
          deps_dir value from fobos, or default if not set
        """
        if self.fobos and self.fobos.has_section("dependencies"):
            if self.fobos.has_option("dependencies", "deps_dir"):
                return self.fobos.get("dependencies", "deps_dir").strip()
        return default

    def rules_list(self, quiet: bool = False) -> None:
        """
        Function for listing defined rules.

        Parameters
        ----------
        quiet : {False}
          less verbose outputs than default
        """
        if self.fobos:
            self.print_n("The fobos file defines the following rules:")
            for rule in self.fobos.sections():
                if rule.startswith("rule-"):
                    if self.fobos.has_option(rule, "help"):
                        helpmsg = self.fobos.get(rule, "help")
                    else:
                        helpmsg = ""
                    self.print_n('  - "' + rule.split("rule-")[1] + '" ' + helpmsg)
                    if self.fobos.has_option(rule, "quiet"):
                        quiet = self.fobos.getboolean(rule, "quiet")
                    for rul in self.fobos.options(rule):
                        if rul.startswith("rule"):
                            if not quiet:
                                self.print_n("       Command => " + self.fobos.get(rule, rul))
        sys.exit(0)
        return

    def rule_execute(self, rule: str, quiet: bool = False, log: bool = False) -> None:
        """
        Function for executing selected rule.

        Parameters
        ----------
        rule : str
          rule name
        quiet : {False}
          less verbose outputs than default
        log : {False}
          bool for activate errors log saving
        """
        if self.fobos:
            self.print_n('Executing rule "' + rule + '"')
            rule_name = "rule-" + rule
            if self.fobos.has_section(rule_name):
                self._get_local_variables()
                self._substitute_local_variables_section(section=rule_name)
                results = []
                quiet = False
                log = False
                if self.fobos.has_option(rule_name, "quiet"):
                    quiet = self.fobos.getboolean(rule_name, "quiet")
                if self.fobos.has_option(rule_name, "log"):
                    log = self.fobos.getboolean(rule_name, "log")
                for rul in self.fobos.options(rule_name):
                    if rul.startswith("rule"):
                        if not quiet:
                            self.print_n("   Command => " + self.fobos.get(rule_name, rul))
                        result = syswork(self.fobos.get(rule_name, rul))
                        results.append(result)
                if log:
                    check_results(results=results, log="rules_errors.log", print_w=self.print_w)
                else:
                    check_results(results=results, print_w=self.print_w)
            else:
                self.print_w('Error: the rule "' + rule + '" is not defined into the fobos file. Defined rules are:')
                self.rules_list(quiet=quiet)
                sys.exit(1)
        return



"""
ParsedFile.py, module definition of Dependency class.
This is a class designed for handling a single parsed file.
"""
# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.

import sys
from typing import Any, Optional, Union

try:
    import functools

    import graphviz as gvz

    __graph__ = functools.partial(gvz.Graph, format="svg")
    __digraph__ = functools.partial(gvz.Digraph, format="svg")
    __styles__ = {
        "graph": {"fontsize": "16", "fontcolor": "black", "bgcolor": "white", "rankdir": "BT"},
        "nodes": {
            "fontname": "Helvetica",
            "shape": "box",
            "fontcolor": "black",
            "color": "black",
            "style": "filled, rounded",
            "fillcolor": "white",
        },
        "edges": {
            "style": "dashed",
            "color": "black",
            "arrowhead": "open",
            "fontname": "Courier",
            "fontsize": "12",
            "fontcolor": "black",
        },
    }
    __graphviz__ = True

    def add_nodes(graph, nodes):
        """
        Auxiliary function for adding nodes to the dependency graph.

        Parameters
        ----------
        graph : __graph__ object
        nodes : nodes to be added to the graph
        """
        for node in nodes:
            if isinstance(node, tuple):
                graph.node(node[0], **node[1])
            else:
                graph.node(node)
        return graph

    def add_edges(graph, edges):
        """
        Auxiliary function for adding edges to the dependency graph.

        Parameters
        ----------
        graph : __graph__ object
        edges : nodes to be added to the graph
        """
        for edge in edges:
            if isinstance(edge[0], tuple):
                graph.edge(*edge[0], **edge[1])
            else:
                graph.edge(*edge)
        return graph

    def apply_styles(graph):
        """
        Auxiliary function for adding styles to the dependency graph.

        Parameters
        ----------
        graph : __graph__ object
        """
        graph.graph_attr.update(("graph" in __styles__ and __styles__["graph"]) or {})
        graph.node_attr.update(("nodes" in __styles__ and __styles__["nodes"]) or {})
        graph.edge_attr.update(("edges" in __styles__ and __styles__["edges"]) or {})
        return graph
except ImportError:
    __graphviz__ = False
import operator
import os
import re
import shlex
from subprocess import STDOUT, check_output

from .Dependency import Dependency
from .Doctest import Doctest
from .utils import traverse_recursive, unique_seq

# definition of regular expressions
__str_apex__ = r"('|" + r'")'
__str_kw_include__ = r"[Ii][Nn][Cc][Ll][Uu][Dd][Ee]"
__str_kw_intrinsic__ = r"[Ii][Nn][Tt][Rr][Ii][Nn][Ss][Ii][Cc]"
__str_kw_iso_fortran_env__ = r"[Ii][Ss][Oo]_[Ff][Oo][Rr][Tt][Rr][Aa][Nn]_[Ee][Nn][Vv]"
__str_kw_iso_c_binding__ = r"[Ii][Ss][Oo]_[Cc]_[Bb][Ii][Nn][Dd][Ii][Nn][Gg]"
__str_kw_ieee_exceptions__ = r"[Ii][Ee][Ee][Ee]_[Ee][Xx][Cc][Ee][Pp][Tt][Ii][Oo][Nn][Ss]"
__str_kw_ieee_arithmetic__ = r"[Ii][Ee][Ee][Ee]_[Aa][Rr][Ii][Tt][Hh][Mm][Ee][Tt][Ii][Cc]"
__str_kw_ieee_features__ = r"[Ii][Ee][Ee][Ee]_[Ff][Ee][Aa][Tt][Uu][Rr][Ee][Ss]"
__str_kw_openacc__ = r"[Oo][Pp][Ee][Nn][Aa][Cc][Cc]"
__str_kw_omp_lib__ = r"[Oo][Mm][Pp]_[Ll][Ii][Bb]"
__str_kw_module__ = r"[Mm][Oo][Dd][Uu][Ll][Ee]"
__str_kw_submodule__ = r"[Ss][Uu][Bb][Mm][Oo][Dd][Uu][Ll][Ee]"
__str_kw_program__ = r"[Pp][Rr][Oo][Gg][Rr][Aa][Mm]"
__str_kw_use__ = r"[Uu][Ss][Ee]"
__str_kw_mpifh__ = r"[Mm][Pp][Ii][Ff]\.[Hh]"
__str_name__ = r"(?P<name>[a-zA-Z][a-zA-Z0-9_]*)"
__str_submodule_ancestor__ = r"(\((?P<submancestor>[a-zA-Z][a-zA-Z0-9_]*)\))"
__str_eol__ = r"(?P<eol>\s*!.*|\s*)?$"
# __str_f95_mod_rename__ = r"(\s*,\s*[a-zA-Z][a-zA-Z0-9_]*\s*=>\s*[a-zA-Z][a-zA-Z0-9_]*)*"
# __str_f95_mod_only__ = r"(\s*,\s*[Oo][Nn][Ll][Yy]\s*:\s*([a-zA-Z][a-zA-Z0-9_]*\s*=>\s*[a-zA-Z][a-zA-Z0-9_]*|[a-zA-Z][a-zA-Z0-9_]*))*"
__str_use_mod__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    +
    # r"(\s*,\s*.*\s*::)*" +  # eventual module nature
    # r"(\s+)" +  # 1 or more white spaces
    r"((\s*,\s*.*\s*::)|(\s*::\s*)|(\s+))"  # eventual module nature
    + __str_name__  # construct name
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_intrinsic__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s*,\s*"
    + __str_kw_intrinsic__
    + r"\s*::.*"  # keyword intrinsic
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_iso_fortran_env__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + __str_kw_iso_fortran_env__  # keyword intrinsic module iso_fortran_env
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_iso_c_binding__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + __str_kw_iso_c_binding__  # keyword intrinsic module iso_c_binding
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_ieee_exceptions__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + __str_kw_ieee_exceptions__  # keyword intrinsic module ieee_exceptions
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_ieee_arithmetic__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + __str_kw_ieee_arithmetic__  # keyword intrinsic module ieee_arithmetic
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_ieee_features__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + __str_kw_ieee_features__  # keyword intrinsic module ieee_features
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_openacc__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + r"(.*)"
    + __str_kw_openacc__
    + r".*"  # keyword intrinsic module openacc
    + r"(?P<mod_eol>(.*))"
)
__str_use_mod_omp_lib__ = (
    r"^(\s*)"  # eventual initial white spaces
    + __str_kw_use__  # keyword "use"
    + r"\s+"
    + r"(.*)"
    + __str_kw_omp_lib__
    + r".*"  # keyword intrinsic module omp_lib
    + r"(?P<mod_eol>(.*))"
)
__str_include__ = (
    r"^(\s*|\#)"  # eventual initial white spaces or "#" character
    + __str_kw_include__  # keyword "include"
    + r"(\s+)"  # 1 or more white spaces
    + __str_apex__  # character "'" or '"'
    + r"(\s*)"  # eventaul white spaces
    + r"(?P<name>.*)"  # name of included file
    + r"(\s*)"  # eventaul white spaces
    + __str_apex__  # character "'" or '"'
    + __str_eol__
)  # eventual eol white space and or comment
__str_module__ = (
    r"^(\s*)"  # eventual initial white spaces
    + r"(?P<scplevel>"
    + __str_kw_module__
    + r")"  # keyword "module"
    + r"(\s+)"  # 1 or more white spaces
    + __str_name__  # construct name
    + __str_eol__
)  # eventual eol white space and or comment
__str_submodule__ = (
    r"^(\s*)"  # eventual initial white spaces
    + r"(?P<scplevel>"
    + __str_kw_submodule__
    + r")"  # keyword "submodule"
    + r"(\s+)"  # 1 or more white spaces
    +
    # r"(\(.*\))" +
    __str_submodule_ancestor__  # (ancestor-module-name [:parent-submodule-name])
    + r"(\s+)"  # 1 or more white spaces
    + __str_name__  # construct name
    + __str_eol__
)  # eventual eol white space and or comment
__str_program__ = (
    r"^(\s*)"  # eventual initial white spaces
    + r"(?P<scplevel>"
    + __str_kw_program__
    + r")"  # keyword "program"
    + r"(\s+)"  # 1 or more white spaces
    + __str_name__  # construct name
    + __str_eol__
)  # eventual eol white space and or comment
__str_mpifh__ = r"(.*)" + __str_kw_mpifh__ + r"(.*)"
__regex_use_mod__ = re.compile(__str_use_mod__)
__regex_include__ = re.compile(__str_include__)
__regex_program__ = re.compile(__str_program__)
__regex_module__ = re.compile(__str_module__)
__regex_submodule__ = re.compile(__str_submodule__)
__regex_use_mod_intrinsic__ = re.compile(__str_use_mod_intrinsic__)
__regex_use_mod_iso_fortran_env__ = re.compile(__str_use_mod_iso_fortran_env__)
__regex_use_mod_iso_c_binding__ = re.compile(__str_use_mod_iso_c_binding__)
__regex_use_mod_ieee_exceptions__ = re.compile(__str_use_mod_ieee_exceptions__)
__regex_use_mod_ieee_arithmetic__ = re.compile(__str_use_mod_ieee_arithmetic__)
__regex_use_mod_ieee_features__ = re.compile(__str_use_mod_ieee_features__)
__regex_use_mod_openacc__ = re.compile(__str_use_mod_openacc__)
__regex_use_mod_omp_lib__ = re.compile(__str_use_mod_omp_lib__)
__regex_use_intrinsic_modules__ = [
    __regex_use_mod_intrinsic__,
    __regex_use_mod_iso_fortran_env__,
    __regex_use_mod_iso_c_binding__,
    __regex_use_mod_ieee_exceptions__,
    __regex_use_mod_ieee_arithmetic__,
    __regex_use_mod_ieee_features__,
    __regex_use_mod_openacc__,
    __regex_use_mod_omp_lib__,
]
__regex_mpifh__ = re.compile(__str_mpifh__)


def openReader(filename: str) -> Any:
    return open(filename, newline="", encoding="utf8")


class ParsedFile:
    """ParsedFile is an object that handles a single parsed file, its attributes and methods."""

    def __init__(
        self,
        name: str,
        program: bool = False,
        module: bool = False,
        submodule: bool = False,
        include: bool = False,
        nomodlib: bool = False,
        to_compile: bool = False,
        output: Optional[str] = None,
        is_doctest: bool = False,
    ) -> None:
        """
        Parameters
        ----------
        name : str
          file name
        program : {False}
          flag for tagging program file
        module : {False}
          flag for tagging module file
        submodule : {False}
          flag for tagging submodule file
        include : {False}
          flag for tagging include file
        nomodlib : {False}
          flag for tagging library of procedures without an enclosing module (old Fortran style)
        to_compile : {False}
          flag for tagging file to be compiled
        output : {None}
          string of output file name

        Attributes
        ----------
        name : str
          file name
        program : bool
          flag for tagging program file
        module : bool
          flag for tagging module file
        submodule : bool
          flag for tagging submodule file
        include : bool
          flag for tagging include file
        nomodlib : bool
          flag for tagging library of procedures without an enclosing module (old Fortran style)
        to_compile : bool
          flag for tagging file to be compiled
        output : str
          string of output file name
        basename : str
          basename of file name
        extension : str
          extension of file name
        timestamp : str
          timestamp of file
        order :
        pfile_dep : list
          list of parsed file dependencies
        doctest : Doctest()
          doctest object
        """
        self.name = name
        self.program = program
        self.module = module
        self.submodule = submodule
        self.include = include
        self.nomodlib = nomodlib
        self.to_compile = to_compile
        self.output = output
        self.is_doctest = is_doctest
        self.basename = os.path.splitext(os.path.basename(self.name))[0]
        self.extension = os.path.splitext(os.path.basename(self.name))[1]
        self.timestamp = os.path.getmtime(self.name)
        self.order = 0
        self.pfile_dep = None
        self.pfile_dep_all = None
        self.module_names = None
        self.submodule_names = None
        self.dependencies = None
        self.doctest = None
        return

    def sort_dependencies(self) -> None:
        """
        Sort dependencies.
        """
        for dep in self.pfile_dep_all:
            for other_dep in self.pfile_dep_all:
                if other_dep != dep:
                    if dep in other_dep.pfile_dep_all:
                        dep.order += 1
        self.pfile_dep_all.sort(key=operator.attrgetter("order"), reverse=True)
        return

    def parse(
        self,
        inc: list[str] = [".INC", ".F", ".FOR", ".FPP", ".F77", ".F90", ".F95", ".F03", ".F08"],
        preprocessor: str = "cpp",
        preproc: str = "",
        include: Union[str, list[str]] = "",
    ) -> None:
        """
        Parse the file creating its the dependencies list and the list of modules names that self eventually contains.

        Parameters
        ----------
        inc : list
          list of extensions of included files
        preprocessor : str
          preprocessor name
        preproc : str
          preprocessor flags
        include : list
          include directories
        """
        self.module_names = []
        self.submodule_names = []
        self.dependencies = []

        if self.extension in inc:
            preprocessor_exist = False
            for path in os.environ["PATH"].split(os.pathsep):
                preprocessor_exist = os.path.exists(os.path.join(path, preprocessor))
                if preprocessor_exist:
                    break
            if preprocessor_exist:
                if preprocessor == "cpp":
                    preprocessor += " -C -w "
                elif preprocessor == "fpp":
                    preprocessor += " -w "
                if preproc is None:
                    preproc = ""
                includes = " "
                if len(include) > 0:
                    includes = " -I" + "-I".join(include) + " "
                source = str(
                    check_output(
                        shlex.split(preprocessor + " " + preproc + includes + self.name),
                        shell=False,
                        stderr=STDOUT,
                        encoding="UTF-8",
                    )
                )
                source = source.replace("\\n", "\n")
            else:
                source = str(openReader(self.name).read())
        else:
            source = str(openReader(self.name).read())

        for line in source.split("\n"):
            matching = re.match(__regex_program__, line)
            if matching:
                self.program = True
            matching = re.match(__regex_module__, line)
            if matching:
                self.module = True
                self.module_names.append(matching.group("name"))
            matching = re.match(__regex_submodule__, line)
            if matching:
                self.submodule = True
                self.submodule_names.append(matching.group("name"))
                dep = Dependency(dtype="module", name=matching.group("submancestor"))
                self.dependencies.append(dep)
            matching = re.match(__regex_use_mod__, line)
            if matching:
                if not any(re.match(regex, line) for regex in __regex_use_intrinsic_modules__):
                    if matching.group("name").lower() != "mpi" and matching.group("name").lower() != "omp_lib":
                        dep = Dependency(dtype="module", name=matching.group("name"))
                        self.dependencies.append(dep)
            matching = re.match(__regex_include__, line)
            if matching:
                if not re.match(__regex_mpifh__, line):
                    dep = Dependency(dtype="include", name=matching.group("name"))
                    self.dependencies.append(dep)

        if self.module:
            self.doctest = Doctest()
            self.doctest.parse(source=source)
            self.doctest.make_volatile_programs()

        if not self.program and not self.module and not self.submodule:
            if os.path.splitext(os.path.basename(self.name))[1] not in inc:
                self.nomodlib = True

    def save_build_log(self, builder: Any) -> None:
        """
        Save a log file containing information about the building options used.

        Parameters
        ----------
        builder : Builder object
          builder used for building self
        """
        with open(os.path.join(builder.build_dir, "build_" + self.basename + ".log"), "w") as log_file:
            log_file.writelines("Hierarchical dependencies list of: " + self.name + "\n")
            for dep in self.pfile_dep:
                log_file.writelines("  " + dep.name + "\n")
                log_file.writelines(dep.str_dependencies(pref="    "))
            log_file.writelines("Complete ordered dependencies list of: " + self.name + "\n")
            for dep in self.pfile_dep_all:
                log_file.writelines("  " + dep.name + "\n")
            log_file.writelines(builder.verbose())
        return

    def save_dep_graph(self) -> None:
        """
        Save dependency graph.
        """
        if __graphviz__:
            depgraph = __digraph__()
            depgraph = add_nodes(depgraph, [self.name])
            depgraph.graph_attr["label"] = "Dependencies graph of " + self.name
            for dep in self.pfile_dep:
                depgraph = add_nodes(depgraph, [dep.name])
                depgraph = add_edges(depgraph, [(self.name, dep.name)])
                for subdep in dep.str_dependencies().split("\n"):
                    if subdep != "":
                        depgraph = add_nodes(depgraph, [subdep])
                        depgraph = add_edges(depgraph, [(dep.name, subdep)])
            depgraph = apply_styles(depgraph)
            depgraph.render("dependency_graph_" + self.basename)
        else:
            print("Module 'graphviz' not found: saving of dependency graph disabled")
        return

    def gnu_make_rule(self, builder: Any) -> str:
        """
        Return the file compiling rule in GNU Make format

        Parameters
        ----------
        builder : Builder object

        Returns
        -------
        str
          string containing the compiling rule of the file
        """
        string = []
        target = self.basename.lower()
        if not self.include:
            objs = []
            for dep in self.pfile_dep:
                if not dep.include and os.path.splitext(os.path.basename(dep.name))[0].lower() != target:
                    objs.append("\t$(DOBJ)" + os.path.splitext(os.path.basename(dep.name))[0].lower() + ".o")
            deps = [self.name]
            for dep in self.pfile_dep:
                if dep.include:
                    deps.append(dep.name)
            if len(objs) > 0:
                string.append("$(DOBJ)" + target + ".o:" + "".join([" " + dep for dep in deps]) + " \\" + "\n")
                for obj in objs[:-1]:
                    string.append(obj + " \\" + "\n")
                string.append(objs[-1] + "\n")
            else:
                string.append("$(DOBJ)" + target + ".o:" + "".join([" " + dep for dep in deps]) + "\n")
            string.append("\t@echo $(COTEXT)\n")
            string.append("\t@$(FC) $(OPTSC) " + "".join(["-I" + i + " " for i in builder.dinc]) + " $< -o $@\n")
            string.append("\n")
        return "".join(string)

    def str_dependencies(self, pref: str = "") -> str:
        """
        Create a string containing the depencies files list.

        Parameters
        ----------
        pref : {""}
          prefixing string
        """
        str_dep = ""
        for dep in self.pfile_dep:
            str_dep += pref + dep.name + "\n"
        return str_dep

    def obj_dependencies(self, exclude_programs: bool = False) -> list[str]:
        """
        Create a list containing the dependencies object files list.

        Parameters
        ----------
        exclude_programs : {False}
          flag for excluding programs obj from the list
        """
        if exclude_programs:
            return [p.basename + ".o" for p in self.pfile_dep_all if not p.include and not p.program]
        else:
            return [p.basename + ".o" for p in self.pfile_dep_all if not p.include]

    def check_compile(self, obj_dir: str, force_compile: bool = False) -> None:
        """
        Check if self must be compiled.

        Parameters
        ----------
        obj_dir : str
          directory where compiled objects are saved
        force_compile : {False}
          flag for forcing (re-)compiling of all dependency
        """
        if not self.include:
            # verifying if dependencies are up-to-date
            for dep in self.pfile_dep_all:
                if not dep.include:
                    if force_compile:
                        self.to_compile = True
                    else:
                        obj = os.path.join(obj_dir, dep.basename + ".o")
                        # verifying if dep is up-to-date
                        if os.path.exists(obj):
                            if os.path.getmtime(obj) < dep.timestamp:
                                # found a dependency object that is out-of-date, thus self (and dep) must be compiled
                                self.to_compile = True
                        else:
                            # compiled object of a dependency is absent, thus self must be compiled
                            self.to_compile = True
                else:
                    # verifying if dep is newer than self
                    if not os.path.exists(dep.name):
                        print(
                            " Attention: file "
                            + dep.name
                            + " does not exist, but it is a dependency of file "
                            + self.name
                        )
                        sys.exit(1)
                    else:
                        # comparing the include dependency with the self-compiled-object if exist
                        obj = os.path.join(obj_dir, self.basename + ".o")
                        # verifying if dep is up-to-date
                        if os.path.exists(obj):
                            if os.path.getmtime(obj) < os.path.getmtime(dep.name):
                                # found an include that is newer than self-compiled-object, thus self must be compiled
                                self.to_compile = True
            # verifying if self is up-to-date
            if not self.to_compile:
                obj = os.path.join(obj_dir, self.basename + ".o")
                if os.path.exists(obj):
                    if os.path.getmtime(obj) < self.timestamp:
                        # the compiled object is out-of-date, thus self must be compiled
                        self.to_compile = True
                else:
                    # compiled object is absent, thus self must be compiled
                    self.to_compile = True

    def create_pfile_dep_all(self) -> None:
        """
        Create a complete list of all dependencies direct and indirect.
        """
        self.pfile_dep_all = []
        for path in traverse_recursive(self):
            self.pfile_dep_all.append(path[-1])
        self.pfile_dep_all = unique_seq(self.pfile_dep_all)
        return



"""
Doctest.py, module definition of Doctest class.

This is a class designed for performing introspective tests
by means of special docstrings.
"""

# Copyright (C) 2015  Stefano Zaghi
#
# This file is part of FoBiS.py.
#
# FoBiS.py is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# FoBiS.py is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with FoBiS.py. If not, see <http://www.gnu.org/licenses/>.
# from builtins import *
import os
import re

__remodule__ = re.compile(
    r"^(\s*)"  # eventual initial white spaces
    + r"([Mm][Oo][Dd][Uu][Ll][Ee])"  # keyword "module"
    + r"(\s+)"  # 1 or more white spaces
    + r"(?P<name>[a-zA-Z][a-zA-Z0-9_]*)"  # module name
    + r"(?P<eol>\s*!.*|\s*)?$"
)  # eventual eol white space and or comment
__redoctest__ = re.compile(r"(?P<doctest>\!.```fortran(?P<test>.*?)\!.```.*?\!\=\>(?P<result>.*?)\<\<\<)", re.DOTALL)


class Doctest:
    """Doctest is an object that handles introspective doc-tests for Fortran procedures,
    like the Python doctest module does for Python codes.

    The syntax of a FoBiS.py doctest for Fortran procedures is:
    !$```fortran
    !$ ! valid fortran code for testing the procedure
    !$ ! the doctest must end with printing the expected result to the standard output
    !$```
    !=> !expected results <<<

    Note that the character "$" can be any character, likely it should be the "docmark" you
    select for producing the API documentation by means of the great FORD autodocumentation
    Fortran tool.
    """

    def __init__(self):
        """
        Parameters
        ----------
        """
        self.to_test = False
        self.tests = []

    def parse(self, source):
        """Parse source string for doctest occurences.

        Parameters
        ----------
        source : str
          source string with eventual doctests
        """
        modules = []
        for line in source.split("\n"):
            match = re.match(__remodule__, line)
            if match:
                modules.append(match)
        if len(modules) > 0:
            for match in re.finditer(__redoctest__, source):
                lines = match.group("test").split("\n")
                lines = [line.strip()[2:] for line in lines if not line.strip() == ""]
                module_name = None
                for module in modules:
                    if module.start() < match.start():
                        module_name = module.group("name")
                self.tests.append({"module": module_name, "source": lines, "result": match.group("result").strip()})

    def make_volatile_programs(self):
        """Make a 'volatile' program test for each of the parsed doctests."""
        if len(self.tests) > 0:
            self.to_test = True
            for test in self.tests:
                program_source = ["program volatile_doctest"]
                program_source.append("use " + test["module"])
                for line in test["source"]:
                    program_source.append(line)
                program_source.append("endprogram volatile_doctest")
                test["program_source"] = "\n".join(program_source)

    def save_volatile_programs(self, build_dir):
        """Save a 'volatile' program test for each of the parsed doctests.

        Parameters
        ----------
        build_dir : str
          build directory where volatile programs are placed

        Returns
        -------
        doctest_root_dir : str
          directory where doctests are saved
        """
        if len(self.tests) > 0:
            doctest_root_dir = os.path.join(build_dir, "doctests-src")
            for t, test in enumerate(self.tests):
                doctest_dir = os.path.join(doctest_root_dir, test["module"])
                if not os.path.exists(doctest_dir):
                    os.makedirs(doctest_dir)
                doctest_file = os.path.join(doctest_dir, test["module"] + "-doctest-" + str(t + 1) + ".f90")
                with open(doctest_file, "w") as doc:
                    doc.writelines(test["program_source"])
                doctest_file = os.path.join(doctest_dir, test["module"] + "-doctest-" + str(t + 1) + ".result")
                with open(doctest_file, "w") as doc:
                    doc.writelines(test["result"])
            return doctest_root_dir



"""Tests for FoBiS.py Makefile generation."""

import pytest

from tests.helpers import make_makefile


@pytest.mark.parametrize("n", range(1, 3))
def test_makefile(n):
    """Test Makefile generation scenario n."""
    assert make_makefile(f"makefile-test{n}"), f"makefile-test{n} failed"
